import numpy as np


def field_dist_1d_vectorized_ji(wavelength, kx_vector, T1, layer_info_list, period,
                                pol, res_x=20, res_y=20, res_z=20, type_complex=np.complex128):

    k0 = 2 * np.pi / wavelength
    Kx = np.diag(kx_vector / k0)

    field_cell = np.zeros((res_z * len(layer_info_list), res_y, res_x, 3), dtype=type_complex)

    T_layer = T1

    # From the first layer
    for idx_layer, (E_conv_i, q, W, X, a_i, b, d) in enumerate(layer_info_list[::-1]):

        c1 = T_layer[:, None]
        c2 = b @ a_i @ X @ T_layer[:, None]

        Q = np.diag(q)

        if pol == 0:
            V = W @ Q
            EKx = None

        else:
            V = E_conv_i @ W @ Q
            EKx = E_conv_i @ Kx

        for k in range(res_z):
            z = k / res_z * d

            if pol == 0:
                Sy = W @ (diag_exp(-k0 * Q * z) @ c1 + diag_exp(k0 * Q * (z - d)) @ c2)
                Ux = V @ (-diag_exp(-k0 * Q * z) @ c1 + diag_exp(k0 * Q * (z - d)) @ c2)
                C = Kx @ Sy

                x_1d = np.arange(res_x).reshape((1, -1, 1))
                x_1d = -1j * x_1d * period[0] / res_x
                x_2d = np.tile(x_1d, (res_y, 1, 1))
                x_2d = x_2d * kx_vector
                x_2d = x_2d.reshape((res_y, res_x, 1, len(kx_vector)))

                exp_K = np.exp(x_2d)
                exp_K = exp_K.reshape((res_y, res_x, -1))

                Ey = exp_K @ Sy
                Hx = -1j * exp_K @ Ux
                Hz = -1j * exp_K @ C

                val = np.concatenate((Ey, Hx, Hz), axis=-1)

            else:
                Uy = W @ (diag_exp(-k0 * Q * z) @ c1 + diag_exp(k0 * Q * (z - d)) @ c2)
                Sx = V @ (-diag_exp(-k0 * Q * z) @ c1 + diag_exp(k0 * Q * (z - d)) @ c2)

                C = EKx @ Uy  # there is a better option for convergence

                x_1d = np.arange(res_x).reshape((1, -1, 1))
                x_1d = -1j * x_1d * period[0] / res_x
                x_2d = np.tile(x_1d, (res_y, 1, 1))
                x_2d = x_2d * kx_vector
                x_2d = x_2d.reshape((res_y, res_x, 1, len(kx_vector)))

                exp_K = np.exp(x_2d)
                exp_K = exp_K.reshape((res_y, res_x, -1))

                Hy = exp_K @ Uy
                Ex = 1j * exp_K @ Sx
                Ez = -1j * exp_K @ C

                val = np.concatenate((Hy, Ex, Ez), axis=-1)

            field_cell[res_z * idx_layer + k] = val

        T_layer = a_i @ X @ T_layer

    return field_cell


def field_dist_1d_conical_vectorized_ji(wavelength, kx_vector, n_I, theta, phi, T1, layer_info_list, period,
                                        res_x=20, res_y=20, res_z=20, type_complex=np.complex128):

    k0 = 2 * np.pi / wavelength
    ky = k0 * n_I * np.sin(theta) * np.sin(phi)
    Kx = np.diag(kx_vector / k0)

    field_cell = np.zeros((res_z * len(layer_info_list), res_y, res_x, 6), dtype=type_complex)

    T_layer = T1

    big_I = np.eye((len(T1))).astype(type_complex)

    # From the first layer
    for idx_layer, [E_conv_i, q_1, q_2, W_1, W_2, V_11, V_12, V_21, V_22, big_X, big_A_i, big_B, d] \
            in enumerate(layer_info_list[::-1]):

        c = np.block([[big_I], [big_B @ big_A_i @ big_X]]) @ T_layer
        cut = len(c) // 4

        c1_plus = c[0*cut:1*cut]
        c2_plus = c[1*cut:2*cut]
        c1_minus = c[2*cut:3*cut]
        c2_minus = c[3*cut:4*cut]

        big_Q1 = np.diag(q_1)
        big_Q2 = np.diag(q_2)

        for k in range(res_z):
            z = k / res_z * d

            Sx = W_2 @ (diag_exp(-k0 * big_Q2 * z) @ c2_plus + diag_exp(k0 * big_Q2 * (z - d)) @ c2_minus)
            Sy = V_11 @ (diag_exp(-k0 * big_Q1 * z) @ c1_plus + diag_exp(k0 * big_Q1 * (z - d)) @ c1_minus) \
                 + V_12 @ (diag_exp(-k0 * big_Q2 * z) @ c2_plus + diag_exp(k0 * big_Q2 * (z - d)) @ c2_minus)
            Ux = W_1 @ (-diag_exp(-k0 * big_Q1 * z) @ c1_plus + diag_exp(k0 * big_Q1 * (z - d)) @ c1_minus)
            Uy = V_21 @ (-diag_exp(-k0 * big_Q1 * z) @ c1_plus + diag_exp(k0 * big_Q1 * (z - d)) @ c1_minus) \
                 + V_22 @ (-diag_exp(-k0 * big_Q2 * z) @ c2_plus + diag_exp(k0 * big_Q2 * (z - d)) @ c2_minus)
            Sz = -1j * E_conv_i @ (Kx @ Uy - ky * Ux)
            Uz = -1j * (Kx @ Sy - ky * Sx)

            x_1d = np.arange(res_x).reshape((1, -1, 1))
            x_1d = -1j * x_1d * period[0] / res_x
            x_2d = np.tile(x_1d, (res_y, 1, 1))
            x_2d = x_2d * kx_vector
            x_2d = x_2d.reshape((res_y, res_x, 1, len(kx_vector)))

            exp_K = np.exp(x_2d)
            exp_K = exp_K.reshape((res_y, res_x, -1))

            Ex = exp_K @ Sx
            Ey = exp_K @ Sy
            Ez = exp_K @ Sz

            Hx = -1j * exp_K @ Ux
            Hy = -1j * exp_K @ Uy
            Hz = -1j * exp_K @ Uz

            val = np.concatenate((Ex, Ey, Ez, Hx, Hy, Hz), axis=-1)

            field_cell[res_z * idx_layer + k] = val

        T_layer = big_A_i @ big_X @ T_layer

    return field_cell


def field_dist_2d_vectorized_ji(wavelength, kx_vector, n_I, theta, phi, fourier_order_x, fourier_order_y, T1, layer_info_list, period,
                                res_x=20, res_y=20, res_z=20, type_complex=np.complex128):

    k0 = 2 * np.pi / wavelength

    fourier_indices_y = np.arange(-fourier_order_y, fourier_order_y + 1)
    ff_x = fourier_order_x * 2 + 1
    ff_y = fourier_order_y * 2 + 1
    ky_vector = k0 * (n_I * np.sin(theta) * np.sin(phi) + fourier_indices_y * (
            wavelength / period[1])).astype(type_complex)

    Kx = np.diag(np.tile(kx_vector, ff_y).flatten()) / k0
    Ky = np.diag(np.tile(ky_vector.reshape((-1, 1)), ff_x).flatten()) / k0

    field_cell = np.zeros((res_z * len(layer_info_list), res_y, res_x, 6), dtype=type_complex)

    T_layer = T1

    big_I = np.eye((len(T1))).astype(type_complex)

    # From the first layer
    for idx_layer, (E_conv_i, q, W_11, W_12, W_21, W_22, V_11, V_12, V_21, V_22, big_X, big_A_i, big_B, d)\
            in enumerate(layer_info_list[::-1]):

        c = np.block([[big_I], [big_B @ big_A_i @ big_X]]) @ T_layer

        ff = len(c) // 4

        c1_plus = c[0*ff:1*ff]
        c2_plus = c[1*ff:2*ff]
        c1_minus = c[2*ff:3*ff]
        c2_minus = c[3*ff:4*ff]

        q1 = q[:len(q)//2]
        q2 = q[len(q)//2:]
        big_Q1 = np.diag(q1)
        big_Q2 = np.diag(q2)

        for k in range(res_z):
            z = k / res_z * d

            Sx = W_11 @ (diag_exp(-k0 * big_Q1 * z) @ c1_plus + diag_exp(k0 * big_Q1 * (z - d)) @ c1_minus) \
                 + W_12 @ (diag_exp(-k0 * big_Q2 * z) @ c2_plus + diag_exp(k0 * big_Q2 * (z - d)) @ c2_minus)
            Sy = W_21 @ (diag_exp(-k0 * big_Q1 * z) @ c1_plus + diag_exp(k0 * big_Q1 * (z - d)) @ c1_minus) \
                 + W_22 @ (diag_exp(-k0 * big_Q2 * z) @ c2_plus + diag_exp(k0 * big_Q2 * (z - d)) @ c2_minus)
            Ux = V_11 @ (-diag_exp(-k0 * big_Q1 * z) @ c1_plus + diag_exp(k0 * big_Q1 * (z - d)) @ c1_minus) \
                 + V_12 @ (-diag_exp(-k0 * big_Q2 * z) @ c2_plus + diag_exp(k0 * big_Q2 * (z - d)) @ c2_minus)
            Uy = V_21 @ (-diag_exp(-k0 * big_Q1 * z) @ c1_plus + diag_exp(k0 * big_Q1 * (z - d)) @ c1_minus) \
                 + V_22 @ (-diag_exp(-k0 * big_Q2 * z) @ c2_plus + diag_exp(k0 * big_Q2 * (z - d)) @ c2_minus)
            Sz = -1j * E_conv_i @ (Kx @ Uy - Ky @ Ux)
            Uz = -1j * (Kx @ Sy - Ky @ Sx)

            x_1d = np.arange(res_x).reshape((1, -1, 1))
            y_1d = np.arange(res_y).reshape((-1, 1, 1))
            x_1d = -1j * x_1d * period[0] / res_x
            y_1d = -1j * y_1d * period[1] / res_y
            x_2d = np.tile(x_1d, (res_y, 1, 1))
            y_2d = np.tile(y_1d, (1, res_x, 1))
            x_2d = x_2d * kx_vector
            y_2d = y_2d * ky_vector
            x_2d = x_2d.reshape((res_y, res_x, 1, len(kx_vector)))
            y_2d = y_2d.reshape((res_y, res_x, len(ky_vector), 1))

            exp_K = np.exp(x_2d) * np.exp(y_2d)
            exp_K = exp_K.reshape((res_y, res_x, -1))

            Ex = exp_K @ Sx
            Ey = exp_K @ Sy
            Ez = exp_K @ Sz
            Hx = -1j * exp_K @ Ux
            Hy = -1j * exp_K @ Uy
            Hz = -1j * exp_K @ Uz

            val = np.concatenate((Ex, Ey, Ez, Hx, Hy, Hz), axis=-1)
            field_cell[res_z * idx_layer + k] = val

        T_layer = big_A_i @ big_X @ T_layer

    return field_cell


def field_dist_1d_vectorized_kji(wavelength, kx_vector, T1, layer_info_list, period,
                                 pol, res_x=20, res_y=20, res_z=20, type_complex=np.complex128):

    k0 = 2 * np.pi / wavelength
    Kx = np.diag(kx_vector / k0)

    field_cell = np.zeros((res_z * len(layer_info_list), res_y, res_x, 3), dtype=type_complex)

    T_layer = T1

    # From the first layer
    for idx_layer, (E_conv_i, q, W, X, a_i, b, d) in enumerate(layer_info_list[::-1]):
        c1 = T_layer[:, None]
        c2 = b @ a_i @ X @ T_layer[:, None]
        Q = np.diag(q)

        if pol == 0:
            V = W @ Q
            EKx = None
        else:
            V = E_conv_i @ W @ Q
            EKx = E_conv_i @ Kx

        z_1d = np.arange(res_z).reshape((-1, 1, 1)) / res_z * d

        if pol == 0:
            Sy = W @ (diag_exp_batch(-k0 * Q * z_1d) @ c1 + diag_exp_batch(k0 * Q * (z_1d - d)) @ c2)
            Ux = V @ (-diag_exp_batch(-k0 * Q * z_1d) @ c1 + diag_exp_batch(k0 * Q * (z_1d - d)) @ c2)
            C = Kx @ Sy

            x_1d = np.arange(res_x).reshape((1, -1, 1))
            x_1d = -1j * x_1d * period[0] / res_x
            x_2d = np.tile(x_1d, (res_y, 1, 1))
            x_2d = x_2d * kx_vector
            x_2d = x_2d.reshape((res_y, res_x, 1, len(kx_vector)))

            exp_K = np.exp(x_2d)
            exp_K = exp_K.reshape((res_y, res_x, -1))

            Ey = exp_K[:, :, None, :] @ Sy[:, None, None, :, :]
            Hx = -1j * exp_K[:, :, None, :] @ Ux[:, None, None, :, :]
            Hz = -1j * exp_K[:, :, None, :] @ C[:, None, None, :, :]

            val = np.concatenate((Ey.squeeze(-1), Hx.squeeze(-1), Hz.squeeze(-1)), axis=-1)

        else:
            Uy = W @ (diag_exp_batch(-k0 * Q * z_1d) @ c1 + diag_exp_batch(k0 * Q * (z_1d - d)) @ c2)
            Sx = V @ (-diag_exp_batch(-k0 * Q * z_1d) @ c1 + diag_exp_batch(k0 * Q * (z_1d - d)) @ c2)

            C = EKx @ Uy  # there is a better option for convergence

            x_1d = np.arange(res_x).reshape((1, -1, 1))
            x_1d = -1j * x_1d * period[0] / res_x
            x_2d = np.tile(x_1d, (res_y, 1, 1))
            x_2d = x_2d * kx_vector
            x_2d = x_2d.reshape((res_y, res_x, 1, len(kx_vector)))

            exp_K = np.exp(x_2d)
            exp_K = exp_K.reshape((res_y, res_x, -1))

            Hy = exp_K[:, :, None, :] @ Uy[:, None, None, :, :]
            Ex = 1j * exp_K[:, :, None, :] @ Sx[:, None, None, :, :]
            Ez = -1j * exp_K[:, :, None, :] @ C[:, None, None, :, :]

            val = np.concatenate((Hy.squeeze(-1), Ex.squeeze(-1), Ez.squeeze(-1)), axis=-1)

        field_cell[res_z * idx_layer:res_z * (idx_layer + 1)] = val

        T_layer = a_i @ X @ T_layer

    return field_cell


def field_dist_1d_conical_vectorized_kji(wavelength, kx_vector, n_I, theta, phi, T1, layer_info_list, period,
                                         res_x=20, res_y=20, res_z=20, type_complex=np.complex128):

    k0 = 2 * np.pi / wavelength
    ky = k0 * n_I * np.sin(theta) * np.sin(phi)
    Kx = np.diag(kx_vector / k0)

    field_cell = np.zeros((res_z * len(layer_info_list), res_y, res_x, 6), dtype=type_complex)

    T_layer = T1

    big_I = np.eye((len(T1))).astype(type_complex)

    # From the first layer
    for idx_layer, [E_conv_i, q_1, q_2, W_1, W_2, V_11, V_12, V_21, V_22, big_X, big_A_i, big_B, d] \
            in enumerate(layer_info_list[::-1]):

        c = np.block([[big_I], [big_B @ big_A_i @ big_X]]) @ T_layer
        z_1d = np.arange(res_z).reshape((-1, 1, 1)) / res_z * d

        ff = len(c) // 4

        c1_plus = c[0 * ff:1 * ff]
        c2_plus = c[1 * ff:2 * ff]
        c1_minus = c[2 * ff:3 * ff]
        c2_minus = c[3 * ff:4 * ff]

        big_Q1 = np.diag(q_1)
        big_Q2 = np.diag(q_2)

        Sx = W_2 @ (diag_exp_batch(-k0 * big_Q2 * z_1d) @ c2_plus + diag_exp_batch(k0 * big_Q2 * (z_1d - d)) @ c2_minus)
        Sy = V_11 @ (diag_exp_batch(-k0 * big_Q1 * z_1d) @ c1_plus + diag_exp_batch(k0 * big_Q1 * (z_1d - d)) @ c1_minus) \
             + V_12 @ (diag_exp_batch(-k0 * big_Q2 * z_1d) @ c2_plus + diag_exp_batch(k0 * big_Q2 * (z_1d - d)) @ c2_minus)
        Ux = W_1 @ (-diag_exp_batch(-k0 * big_Q1 * z_1d) @ c1_plus + diag_exp_batch(k0 * big_Q1 * (z_1d - d)) @ c1_minus)
        Uy = V_21 @ (-diag_exp_batch(-k0 * big_Q1 * z_1d) @ c1_plus + diag_exp_batch(k0 * big_Q1 * (z_1d - d)) @ c1_minus) \
             + V_22 @ (-diag_exp_batch(-k0 * big_Q2 * z_1d) @ c2_plus + diag_exp_batch(k0 * big_Q2 * (z_1d - d)) @ c2_minus)
        Sz = -1j * E_conv_i @ (Kx @ Uy - ky * Ux)
        Uz = -1j * (Kx @ Sy - ky * Sx)

        x_1d = np.arange(res_x).reshape((1, -1, 1))
        x_1d = -1j * x_1d * period[0] / res_x
        x_2d = np.tile(x_1d, (res_y, 1, 1))
        x_2d = x_2d * kx_vector
        x_2d = x_2d.reshape((res_y, res_x, 1, len(kx_vector)))

        exp_K = np.exp(x_2d)
        exp_K = exp_K.reshape((res_y, res_x, -1))

        Ex = exp_K[:, :, None, :] @ Sx[:, None, None, :, :]
        Ey = exp_K[:, :, None, :] @ Sy[:, None, None, :, :]
        Ez = exp_K[:, :, None, :] @ Sz[:, None, None, :, :]

        Hx = -1j * exp_K[:, :, None, :] @ Ux[:, None, None, :, :]
        Hy = -1j * exp_K[:, :, None, :] @ Uy[:, None, None, :, :]
        Hz = -1j * exp_K[:, :, None, :] @ Uz[:, None, None, :, :]

        val = np.concatenate(
            (Ex.squeeze(-1), Ey.squeeze(-1), Ez.squeeze(-1), Hx.squeeze(-1), Hy.squeeze(-1), Hz.squeeze(-1)), -1)
        field_cell[res_z * idx_layer:res_z * (idx_layer + 1)] = val
        T_layer = big_A_i @ big_X @ T_layer

    return field_cell


def field_dist_2d_vectorized_kji(wavelength, kx_vector, n_I, theta, phi, fourier_order_x, fourier_order_y, T1, layer_info_list, period,
                                 res_x=20, res_y=20, res_z=20, type_complex=np.complex128):

    k0 = 2 * np.pi / wavelength

    fourier_indices_y = np.arange(-fourier_order_y, fourier_order_y + 1)
    ff_x = fourier_order_x * 2 + 1
    ff_y = fourier_order_y * 2 + 1
    ky_vector = k0 * (n_I * np.sin(theta) * np.sin(phi) + fourier_indices_y * (
            wavelength / period[1])).astype(type_complex)

    Kx = np.diag(np.tile(kx_vector, ff_y).flatten()) / k0
    Ky = np.diag(np.tile(ky_vector.reshape((-1, 1)), ff_x).flatten()) / k0

    field_cell = np.zeros((res_z * len(layer_info_list), res_y, res_x, 6), dtype=type_complex)

    T_layer = T1

    big_I = np.eye((len(T1))).astype(type_complex)

    # From the first layer
    for idx_layer, (E_conv_i, q, W_11, W_12, W_21, W_22, V_11, V_12, V_21, V_22, big_X, big_A_i, big_B, d)\
            in enumerate(layer_info_list[::-1]):

        c = np.block([[big_I], [big_B @ big_A_i @ big_X]]) @ T_layer
        z_1d = np.arange(res_z).reshape((-1, 1, 1)) / res_z * d

        ff = len(c) // 4

        c1_plus = c[0 * ff:1 * ff]
        c2_plus = c[1 * ff:2 * ff]
        c1_minus = c[2 * ff:3 * ff]
        c2_minus = c[3 * ff:4 * ff]

        q1 = q[:len(q) // 2]
        q2 = q[len(q) // 2:]
        big_Q1 = np.diag(q1)
        big_Q2 = np.diag(q2)

        Sx = W_11 @ (diag_exp_batch(-k0 * big_Q1 * z_1d) @ c1_plus + diag_exp_batch(k0 * big_Q1 * (z_1d - d)) @ c1_minus) \
              + W_12 @ (diag_exp_batch(-k0 * big_Q2 * z_1d) @ c2_plus + diag_exp_batch(k0 * big_Q2 * (z_1d - d)) @ c2_minus)
        Sy = W_21 @ (diag_exp_batch(-k0 * big_Q1 * z_1d) @ c1_plus + diag_exp_batch(k0 * big_Q1 * (z_1d - d)) @ c1_minus) \
              + W_22 @ (diag_exp_batch(-k0 * big_Q2 * z_1d) @ c2_plus + diag_exp_batch(k0 * big_Q2 * (z_1d - d)) @ c2_minus)
        Ux = V_11 @ (-diag_exp_batch(-k0 * big_Q1 * z_1d) @ c1_plus + diag_exp_batch(k0 * big_Q1 * (z_1d - d)) @ c1_minus) \
              + V_12 @ (-diag_exp_batch(-k0 * big_Q2 * z_1d) @ c2_plus + diag_exp_batch(k0 * big_Q2 * (z_1d - d)) @ c2_minus)
        Uy = V_21 @ (-diag_exp_batch(-k0 * big_Q1 * z_1d) @ c1_plus + diag_exp_batch(k0 * big_Q1 * (z_1d - d)) @ c1_minus) \
              + V_22 @ (-diag_exp_batch(-k0 * big_Q2 * z_1d) @ c2_plus + diag_exp_batch(k0 * big_Q2 * (z_1d - d)) @ c2_minus)
        Sz = -1j * E_conv_i @ (Kx @ Uy - Ky @ Ux)
        Uz = -1j * (Kx @ Sy - Ky @ Sx)

        x_1d = np.arange(res_x).reshape((1, -1, 1))
        y_1d = np.arange(res_y).reshape((-1, 1, 1))
        x_1d = -1j * x_1d * period[0] / res_x
        y_1d = -1j * y_1d * period[1] / res_y
        x_2d = np.tile(x_1d, (res_y, 1, 1))
        y_2d = np.tile(y_1d, (1, res_x, 1))
        x_2d = x_2d * kx_vector
        y_2d = y_2d * ky_vector
        x_2d = x_2d.reshape((res_y, res_x, 1, len(kx_vector)))
        y_2d = y_2d.reshape((res_y, res_x, len(ky_vector), 1))

        exp_K = np.exp(x_2d) * np.exp(y_2d)
        exp_K = exp_K.reshape((res_y, res_x, -1))

        Ex = exp_K[:, :, None, :] @ Sx[:, None, None, :, :]
        Ey = exp_K[:, :, None, :] @ Sy[:, None, None, :, :]
        Ez = exp_K[:, :, None, :] @ Sz[:, None, None, :, :]
        Hx = -1j * exp_K[:, :, None, :] @ Ux[:, None, None, :, :]
        Hy = -1j * exp_K[:, :, None, :] @ Uy[:, None, None, :, :]
        Hz = -1j * exp_K[:, :, None, :] @ Uz[:, None, None, :, :]

        val = np.concatenate(
            (Ex.squeeze(-1), Ey.squeeze(-1), Ez.squeeze(-1), Hx.squeeze(-1), Hy.squeeze(-1), Hz.squeeze(-1)), -1)
        field_cell[res_z * idx_layer:res_z * (idx_layer + 1)] = val

        T_layer = big_A_i @ big_X @ T_layer

    return field_cell


def field_dist_1d_vanilla(wavelength, kx_vector, T1, layer_info_list, period, pol, res_x=20, res_y=20, res_z=20,
                          type_complex=np.complex128):

    k0 = 2 * np.pi / wavelength
    Kx = np.diag(kx_vector / k0)

    field_cell = np.zeros((res_z * len(layer_info_list), res_y, res_x, 3), dtype=type_complex)

    T_layer = T1

    # From the first layer
    for idx_layer, (E_conv_i, q, W, X, a_i, b, d) in enumerate(layer_info_list[::-1]):
        c1 = T_layer[:, None]
        c2 = b @ a_i @ X @ T_layer[:, None]
        Q = np.diag(q)

        if pol == 0:
            V = W @ Q
            EKx = None
        else:
            V = E_conv_i @ W @ Q
            EKx = E_conv_i @ Kx

        for k in range(res_z):
            z = k / res_z * d

            if pol == 0:  # TE
                Sy = W @ (diag_exp(-k0 * Q * z) @ c1 + diag_exp(k0 * Q * (z - d)) @ c2)
                Ux = V @ (-diag_exp(-k0 * Q * z) @ c1 + diag_exp(k0 * Q * (z - d)) @ c2)
                f_here = (-1j) * Kx @ Sy

                for j in range(res_y):
                    for i in range(res_x):
                        x = i * period[0] / res_x

                        Ey = Sy.T @ np.exp(-1j * kx_vector.reshape((-1, 1)) * x)
                        Hx = -1j * Ux.T @ np.exp(-1j * kx_vector.reshape((-1, 1)) * x)
                        Hz = f_here.T @ np.exp(-1j * kx_vector.reshape((-1, 1)) * x)

                        field_cell[res_z * idx_layer + k, j, i] = [Ey[0, 0], Hx[0, 0], Hz[0, 0]]
            else:  # TM
                Uy = W @ (diag_exp(-k0 * Q * z) @ c1 + diag_exp(k0 * Q * (z - d)) @ c2)
                Sx = V @ (-diag_exp(-k0 * Q * z) @ c1 + diag_exp(k0 * Q * (z - d)) @ c2)

                f_here = (-1j) * EKx @ Uy  # there is a better option for convergence

                for j in range(res_y):
                    for i in range(res_x):
                        x = i * period[0] / res_x

                        Hy = Uy.T @ np.exp(-1j * kx_vector.reshape((-1, 1)) * x)
                        Ex = 1j * Sx.T @ np.exp(-1j * kx_vector.reshape((-1, 1)) * x)
                        Ez = f_here.T @ np.exp(-1j * kx_vector.reshape((-1, 1)) * x)

                        field_cell[res_z * idx_layer + k, j, i] = [Hy[0, 0], Ex[0, 0], Ez[0, 0]]

        T_layer = a_i @ X @ T_layer

    return field_cell


def field_dist_1d_conical_vanilla(wavelength, kx_vector, n_I, theta, phi, T1, layer_info_list, period,
                                  res_x=20, res_y=20, res_z=20, type_complex=np.complex128):

    k0 = 2 * np.pi / wavelength
    ky = k0 * n_I * np.sin(theta) * np.sin(phi)
    Kx = np.diag(kx_vector / k0)

    field_cell = np.zeros((res_z * len(layer_info_list), res_y, res_x, 6), dtype=type_complex)

    T_layer = T1
    big_I = np.eye((len(T1)), dtype=type_complex)

    # From the first layer
    for idx_layer, [E_conv_i, q_1, q_2, W_1, W_2, V_11, V_12, V_21, V_22, big_X, big_A_i, big_B, d] \
            in enumerate(layer_info_list[::-1]):

        c = np.block([[big_I], [big_B @ big_A_i @ big_X]]) @ T_layer

        cut = len(c) // 4

        c1_plus = c[0*cut:1*cut]
        c2_plus = c[1*cut:2*cut]
        c1_minus = c[2*cut:3*cut]
        c2_minus = c[3*cut:4*cut]

        big_Q1 = np.diag(q_1)
        big_Q2 = np.diag(q_2)

        for k in range(res_z):
            z = k / res_z * d

            Sx = W_2 @ (diag_exp(-k0 * big_Q2 * z) @ c2_plus + diag_exp(k0 * big_Q2 * (z - d)) @ c2_minus)
            Sy = V_11 @ (diag_exp(-k0 * big_Q1 * z) @ c1_plus + diag_exp(k0 * big_Q1 * (z - d)) @ c1_minus) \
                 + V_12 @ (diag_exp(-k0 * big_Q2 * z) @ c2_plus + diag_exp(k0 * big_Q2 * (z - d)) @ c2_minus)
            Ux = W_1 @ (-diag_exp(-k0 * big_Q1 * z) @ c1_plus + diag_exp(k0 * big_Q1 * (z - d)) @ c1_minus)
            Uy = V_21 @ (-diag_exp(-k0 * big_Q1 * z) @ c1_plus + diag_exp(k0 * big_Q1 * (z - d)) @ c1_minus) \
                 + V_22 @ (-diag_exp(-k0 * big_Q2 * z) @ c2_plus + diag_exp(k0 * big_Q2 * (z - d)) @ c2_minus)
            Sz = -1j * E_conv_i @ (Kx @ Uy - ky * Ux)
            Uz = -1j * (Kx @ Sy - ky * Sx)

            for j in range(res_y):
                for i in range(res_x):
                    x = i * period[0] / res_x
                    exp_K = np.exp(-1j*kx_vector.reshape((-1, 1)) * x)
                    # exp_K = exp_K.flatten()

                    Ex = Sx.T @ exp_K
                    Ey = Sy.T @ exp_K
                    Ez = Sz.T @ exp_K
                    Hx = -1j * Ux.T @ exp_K
                    Hy = -1j * Uy.T @ exp_K
                    Hz = -1j * Uz.T @ exp_K

                    field_cell[res_z * idx_layer + k, j, i] = [Ex[0, 0], Ey[0, 0], Ez[0, 0], Hx[0, 0], Hy[0, 0], Hz[0, 0]]

        T_layer = big_A_i @ big_X @ T_layer

    return field_cell


def field_dist_2d_vanilla(wavelength, kx_vector, n_I, theta, phi, fourier_order_x, fourier_order_y, T1, layer_info_list,
                          period, res_x=20, res_y=20, res_z=20,
                          type_complex=np.complex128):

    k0 = 2 * np.pi / wavelength

    fourier_indices_y = np.arange(-fourier_order_y, fourier_order_y + 1)
    ff_x = fourier_order_x * 2 + 1
    ff_y = fourier_order_y * 2 + 1

    ky_vector = k0 * (n_I * np.sin(theta) * np.sin(phi) + fourier_indices_y * (
            wavelength / period[1])).astype(type_complex)

    Kx = np.diag(np.tile(kx_vector, ff_y).flatten()) / k0
    Ky = np.diag(np.tile(ky_vector.reshape((-1, 1)), ff_x).flatten()) / k0

    field_cell = np.zeros((res_z * len(layer_info_list), res_y, res_x, 6), dtype=type_complex)

    T_layer = T1
    big_I = np.eye((len(T1)), dtype=type_complex)

    # From the first layer
    for idx_layer, (E_conv_i, q, W_11, W_12, W_21, W_22, V_11, V_12, V_21, V_22, big_X, big_A_i, big_B, d)\
            in enumerate(layer_info_list[::-1]):

        c = np.block([[big_I], [big_B @ big_A_i @ big_X]]) @ T_layer

        ff = len(c) // 4

        c1_plus = c[0*ff:1*ff]
        c2_plus = c[1*ff:2*ff]
        c1_minus = c[2*ff:3*ff]
        c2_minus = c[3*ff:4*ff]

        q1 = q[:len(q)//2]
        q2 = q[len(q)//2:]
        big_Q1 = np.diag(q1)
        big_Q2 = np.diag(q2)

        for k in range(res_z):
            z = k / res_z * d

            Sx = W_11 @ (diag_exp(-k0 * big_Q1 * z) @ c1_plus + diag_exp(k0 * big_Q1 * (z - d)) @ c1_minus) \
                 + W_12 @ (diag_exp(-k0 * big_Q2 * z) @ c2_plus + diag_exp(k0 * big_Q2 * (z - d)) @ c2_minus)
            Sy = W_21 @ (diag_exp(-k0 * big_Q1 * z) @ c1_plus + diag_exp(k0 * big_Q1 * (z - d)) @ c1_minus) \
                 + W_22 @ (diag_exp(-k0 * big_Q2 * z) @ c2_plus + diag_exp(k0 * big_Q2 * (z - d)) @ c2_minus)
            Ux = V_11 @ (-diag_exp(-k0 * big_Q1 * z) @ c1_plus + diag_exp(k0 * big_Q1 * (z - d)) @ c1_minus) \
                 + V_12 @ (-diag_exp(-k0 * big_Q2 * z) @ c2_plus + diag_exp(k0 * big_Q2 * (z - d)) @ c2_minus)
            Uy = V_21 @ (-diag_exp(-k0 * big_Q1 * z) @ c1_plus + diag_exp(k0 * big_Q1 * (z - d)) @ c1_minus) \
                 + V_22 @ (-diag_exp(-k0 * big_Q2 * z) @ c2_plus + diag_exp(k0 * big_Q2 * (z - d)) @ c2_minus)
            Sz = -1j * E_conv_i @ (Kx @ Uy - Ky @ Ux)
            Uz = -1j * (Kx @ Sy - Ky @ Sx)

            for j in range(res_y):
                y = j * period[1] / res_y

                for i in range(res_x):
                    x = i * period[0] / res_x

                    exp_K = np.exp(-1j*kx_vector.reshape((1, -1)) * x) * np.exp(-1j*ky_vector.reshape((-1, 1)) * y)
                    exp_K = exp_K.flatten()

                    Ex = Sx.T @ exp_K
                    Ey = Sy.T @ exp_K
                    Ez = Sz.T @ exp_K
                    Hx = -1j * Ux.T @ exp_K
                    Hy = -1j * Uy.T @ exp_K
                    Hz = -1j * Uz.T @ exp_K

                    field_cell[res_z * idx_layer + k, j, i] = [Ex[0], Ey[0], Ez[0], Hx[0], Hy[0], Hz[0]]
        T_layer = big_A_i @ big_X @ T_layer

    return field_cell


def field_plot(field_cell, pol=0, plot_indices=(1, 1, 1, 1, 1, 1), y_slice=0, z_slice=-1, zx=True, yx=True):
    try:
        import matplotlib.pyplot as plt
    except (ImportError, ModuleNotFoundError) as e:
        print(e)
        print('To use cal_field(), please install matplotlib')
        raise e

    if field_cell.shape[-1] == 6:  # 2D grating
        title = ['2D Ex', '2D Ey', '2D Ez', '2D Hx', '2D Hy', '2D Hz', ]
    else:  # 1D grating
        if pol == 0:  # TE
            title = ['1D Ey', '1D Hx', '1D Hz', ]
        else:  # TM
            title = ['1D Hy', '1D Ex', '1D Ez', ]

    if zx:
        for idx in range(len(title)):
            if plot_indices[idx]:
                plt.imshow((abs(field_cell[:, y_slice, :, idx]) ** 2), cmap='jet', aspect='auto')
                # plt.clim(0, 2)  # identical to caxis([-4,4]) in MATLAB
                plt.colorbar()
                plt.title(f'{title[idx]}, Side View')
                plt.xlabel('X')
                plt.ylabel('Z')
                plt.show()
    if yx:
        for idx in range(len(title)):
            if plot_indices[idx]:
                plt.imshow((abs(field_cell[z_slice, :, :, idx]) ** 2), cmap='jet', aspect='auto')
                # plt.clim(0, 3.5)  # identical to caxis([-4,4]) in MATLAB
                plt.colorbar()
                plt.title(f'{title[idx]}, Top View')
                plt.xlabel('X')
                plt.ylabel('Y')
                plt.show()


def diag_exp(x):
    return np.diag(np.exp(np.diag(x)))


def diag_exp_batch(x):
    res = np.zeros(x.shape).astype(x.dtype)
    ix = np.arange(x.shape[-1])
    res[:, ix, ix] = np.exp(x[:, ix, ix])
    return res



import time
from functools import partial

import jax

import numpy as np
import jax.numpy as jnp

from ._base import _BaseRCWA
from .convolution_matrix import to_conv_mat_raster_discrete, to_conv_mat_raster_continuous, to_conv_mat_vector
from .field_distribution import field_dist_1d_vectorized_ji, field_dist_1d_conical_vectorized_ji, \
    field_dist_2d_vectorized_ji, field_plot, \
    field_dist_1d_vectorized_kji, field_dist_1d_conical_vectorized_kji, field_dist_1d_vanilla, \
    field_dist_1d_conical_vanilla, field_dist_2d_vanilla, field_dist_2d_vectorized_kji


class RCWAJax(_BaseRCWA):
    def __init__(self,
                 n_I=1.,
                 n_II=1.,
                 theta=0.,
                 phi=0.,
                 period=(100., 100.),
                 wavelength=900.,
                 ucell=None,
                 ucell_info_list=None,
                 thickness=(0., ),
                 backend=1,
                 grating_type=0,
                 pol=0.,
                 fourier_order=(2, 0),
                 ucell_materials=None,
                 algo='TMM',
                 perturbation=1E-20,
                 device='cpu',
                 type_complex=jnp.complex128,
                 fft_type=0,
                 improve_dft=True,
                 **kwargs,
                 ):

        super().__init__(grating_type=grating_type, n_I=n_I, n_II=n_II, theta=theta, phi=phi, pol=pol,
                         fourier_order=fourier_order, period=period, wavelength=wavelength,
                         thickness=thickness, algo=algo, perturbation=perturbation,
                         device=device, type_complex=type_complex)

        self.ucell = ucell
        self.ucell_materials = ucell_materials
        self.ucell_info_list = ucell_info_list

        self.backend = backend
        self.fft_type = fft_type
        self.improve_dft = improve_dft

        self.layer_info_list = []

    def _tree_flatten(self):
        children = (self.n_I, self.n_II, self.theta, self.phi, self.psi,
                    self.period, self.wavelength, self.ucell, self.ucell_info_list, self.thickness)
        aux_data = {
            'backend': self.backend,
            'grating_type': self.grating_type,
            'pol': self.pol,
            'fourier_order': self.fourier_order,
            'ucell_materials': self.ucell_materials,
            'algo': self.algo,
            'perturbation': self.perturbation,
            'device': self.device,
            'type_complex': self.type_complex,
            'fft_type': self.fft_type,
        }

        return children, aux_data

    @classmethod
    def _tree_unflatten(cls, aux_data, children):
        return cls(*children, **aux_data)

    @property
    def ucell(self):
        return self._ucell

    @ucell.setter
    def ucell(self, ucell):
        if isinstance(ucell, jnp.ndarray):
            if ucell.dtype in (jnp.float64, jnp.float32, jnp.int64, jnp.int32):
                dtype = self.type_float
            elif ucell.dtype in (jnp.complex128, jnp.complex64):
                dtype = self.type_complex
            else:
                raise ValueError
            self._ucell = ucell.astype(dtype)
        elif isinstance(ucell, np.ndarray):
            if ucell.dtype in (np.int64, np.float64, np.int32, np.float32):
                dtype = self.type_float
            elif ucell.dtype in (np.complex128, np.complex64):
                dtype = self.type_complex
            else:
                raise ValueError
            self._ucell = jnp.array(ucell, dtype=dtype)
        elif ucell is None:
            self._ucell = ucell
        else:
            raise ValueError

    def _solve(self, wavelength, e_conv_all, o_e_conv_all):
        self.kx_vector = self.get_kx_vector(wavelength)

        if self.grating_type == 0:
            de_ri, de_ti, layer_info_list, T1 = self.solve_1d(wavelength, e_conv_all, o_e_conv_all)
        elif self.grating_type == 1:
            de_ri, de_ti, layer_info_list, T1 = self.solve_1d_conical(wavelength, e_conv_all, o_e_conv_all)
        elif self.grating_type == 2:
            de_ri, de_ti, layer_info_list, T1 = self.solve_2d(wavelength, e_conv_all, o_e_conv_all)
        else:
            raise ValueError

        return de_ri, de_ti, layer_info_list, T1, self.kx_vector

    @_BaseRCWA.jax_device_set
    def solve(self, wavelength, e_conv_all, o_e_conv_all):
        de_ri, de_ti, layer_info_list, T1, kx_vector = jax.jit(self._solve)(wavelength, e_conv_all, o_e_conv_all)

        self.layer_info_list = layer_info_list
        self.T1 = T1
        self.kx_vector = kx_vector

        return de_ri, de_ti

    def _conv_solve(self):

        if self.fft_type == 0:
            E_conv_all, o_E_conv_all = to_conv_mat_raster_discrete(self.ucell, self.fourier_order[0], self.fourier_order[1],
                                                                   type_complex=self.type_complex, improve_dft=self.improve_dft)
        elif self.fft_type == 1:
            E_conv_all, o_E_conv_all = to_conv_mat_raster_continuous(self.ucell, self.fourier_order[0], self.fourier_order[1],
                                                                     type_complex=self.type_complex)
        elif self.fft_type == 2:
            E_conv_all, o_E_conv_all = to_conv_mat_vector(self.ucell_info_list,
                                                          self.fourier_order[0], self.fourier_order[1],
                                                          type_complex=self.type_complex)
        else:
            raise ValueError

        de_ri, de_ti, layer_info_list, T1, kx_vector = self._solve(self.wavelength, E_conv_all, o_E_conv_all)
        return de_ri, de_ti, layer_info_list, T1, kx_vector

    @jax.jit
    def _conv_solve_jit(self):
        return self._conv_solve()

    @_BaseRCWA.jax_device_set
    def conv_solve(self, **kwargs):
        [setattr(self, k, v) for k, v in kwargs.items()]  # needed for optimization
        if self.fft_type == 1:
            # print('CFT (fft_type=1) is not supported for jit-compilation. Using non-jit-compiled method.')
            de_ri, de_ti, layer_info_list, T1, kx_vector = self._conv_solve()

        else:
            de_ri, de_ti, layer_info_list, T1, kx_vector = self._conv_solve_jit()

        self.layer_info_list = layer_info_list
        self.T1 = T1
        self.kx_vector = kx_vector

        return de_ri, de_ti

    @_BaseRCWA.jax_device_set
    def calculate_field(self, res_x=20, res_y=20, res_z=20, field_algo=2):

        if self.grating_type == 0:
            res_y = 1
            if field_algo == 0:
                field_cell = field_dist_1d_vanilla(self.wavelength, self.kx_vector,
                                                   self.T1, self.layer_info_list, self.period, self.pol,
                                                   res_x=res_x, res_y=res_y, res_z=res_z,
                                                   type_complex=self.type_complex)
            elif field_algo == 1:
                field_cell = field_dist_1d_vectorized_ji(self.wavelength, self.kx_vector, self.T1, self.layer_info_list,
                                                         self.period, self.pol, res_x=res_x, res_y=res_y, res_z=res_z,
                                                         type_complex=self.type_complex, type_float=self.type_float)
            elif field_algo == 2:
                field_cell = field_dist_1d_vectorized_kji(self.wavelength, self.kx_vector, self.T1,
                                                          self.layer_info_list, self.period, self.pol,
                                                          res_x=res_x, res_y=res_y, res_z=res_z,
                                                          type_complex=self.type_complex, type_float=self.type_float)
            else:
                raise ValueError

        elif self.grating_type == 1:
            res_y = 1
            if field_algo == 0:
                field_cell = field_dist_1d_conical_vanilla(self.wavelength, self.kx_vector, self.n_I, self.theta,
                                                           self.phi, self.T1, self.layer_info_list, self.period,
                                                           res_x=res_x, res_y=res_y, res_z=res_z,
                                                           type_complex=self.type_complex)
            elif field_algo == 1:
                field_cell = field_dist_1d_conical_vectorized_ji(self.wavelength, self.kx_vector, self.n_I, self.theta,
                                                                 self.phi, self.T1, self.layer_info_list, self.period,
                                                                 res_x=res_x, res_y=res_y, res_z=res_z,
                                                                 type_complex=self.type_complex, type_float=self.type_float)
            elif field_algo == 2:
                field_cell = field_dist_1d_conical_vectorized_kji(self.wavelength, self.kx_vector, self.n_I, self.theta,
                                                                  self.phi, self.T1, self.layer_info_list, self.period,
                                                                  res_x=res_x, res_y=res_y, res_z=res_z,
                                                                  type_complex=self.type_complex, type_float=self.type_float)
            else:
                raise ValueError

        elif self.grating_type == 2:

            if field_algo == 0:
                field_cell = field_dist_2d_vanilla(self.wavelength, self.kx_vector, self.n_I, self.theta, self.phi,
                                                   *self.fourier_order, self.T1, self.layer_info_list, self.period,
                                                   res_x=res_x, res_y=res_y, res_z=res_z,
                                                   type_complex=self.type_complex)
            elif field_algo == 1:
                field_cell = field_dist_2d_vectorized_ji(self.wavelength, self.kx_vector, self.n_I, self.theta,
                                                         self.phi, *self.fourier_order, self.T1, self.layer_info_list,
                                                         self.period, res_x=res_x, res_y=res_y, res_z=res_z,
                                                         type_complex=self.type_complex, type_float=self.type_float)
            elif field_algo == 2:
                field_cell = field_dist_2d_vectorized_kji(self.wavelength, self.kx_vector, self.n_I, self.theta,
                                                          self.phi, *self.fourier_order, self.T1, self.layer_info_list,
                                                          self.period, res_x=res_x, res_y=res_y, res_z=res_z,
                                                          type_complex=self.type_complex, type_float=self.type_float)
            else:
                raise ValueError
        else:
            raise ValueError

        return field_cell

    def field_plot(self, field_cell):
        field_plot(field_cell, self.pol)

    @_BaseRCWA.jax_device_set
    def calculate_field_all(self, res_x=20, res_y=20, res_z=20):
        t0 = time.time()
        field_cell0 = self.calculate_field(res_x=res_x, res_y=res_y, res_z=res_z, field_algo=0)
        print('no vector', time.time() - t0)
        t0 = time.time()
        field_cell1 = self.calculate_field(res_x=res_x, res_y=res_y, res_z=res_z, field_algo=1)
        print('ji vector', time.time() - t0)
        t0 = time.time()
        field_cell2 = self.calculate_field(res_x=res_x, res_y=res_y, res_z=res_z, field_algo=2)
        print('kji vector', time.time() - t0)

        print('gap(1-0): ', jnp.linalg.norm(field_cell1 - field_cell0))
        print('gap(2-1): ', jnp.linalg.norm(field_cell2 - field_cell1))
        print('gap(0-2): ', jnp.linalg.norm(field_cell0 - field_cell2))

        return field_cell0, field_cell1, field_cell2

    @partial(jax.jit, static_argnums=(1, 2, 3, 4))
    @_BaseRCWA.jax_device_set
    def conv_solve_field(self, res_x=20, res_y=20, res_z=20, field_algo=2, **kwargs):
        [setattr(self, k, v) for k, v in kwargs.items()]  # needed for optimization

        if self.fft_type == 1:
            print('CFT (fft_type=1) is not supported with JAX jit-compilation. Use conv_solve_field_no_jit.')
            return None, None, None

        de_ri, de_ti, _, _, _ = self._conv_solve()
        field_cell = self.calculate_field(res_x, res_y, res_z, field_algo=field_algo)
        return de_ri, de_ti, field_cell

    @_BaseRCWA.jax_device_set
    def conv_solve_field_no_jit(self, res_x=20, res_y=20, res_z=20, field_algo=2):
        de_ri, de_ti, _, _, _ = self._conv_solve()
        field_cell = self.calculate_field(res_x, res_y, res_z, field_algo=field_algo)
        return de_ri, de_ti, field_cell

    def run_ucell_vmap(self, ucell_list):
        """
        under development
        """

        de_ri, de_ti = jax.vmap(self.conv_solve)(ucell_list)

        return de_ri, de_ti

    def run_ucell_pmap(self, ucell_list):
        """
        under development
        """

        de_ri, de_ti = jax.pmap(self.conv_solve)(ucell_list)
        # de_ri.block_until_ready()
        # de_ti.block_until_ready()

        de_ri = jnp.array(de_ri)
        de_ti = jnp.array(de_ti)
        return de_ri, de_ti


    # def generate_spectrum(self, wl_list):
    #     ucell = deepcopy(self.ucell)
    #     spectrum_ri_pmap = jnp.zeros(wl_list.shape)
    #     spectrum_ti_pmap = jnp.zeros(wl_list.shape)
    #     for i, wavelength in enumerate(range(counter)):
    #         b = i * num_device
    #         de_ri_pmap, de_ti_pmap = jax.pmap(loop_wavelength)(wavelength_list[b:b + num_device],
    #                                                            mat_pmtvy_interp[b:b + num_device])
    #
    #         spectrum_ri_pmap[b:b + num_device] = de_ri_pmap.sum(axis=(1, 2))
    #         spectrum_ti_pmap[b:b + num_device] = de_ti_pmap.sum(axis=(1, 2))
    #
    #     return spectrum_ri_pmap, spectrum_ti_pmap



# from jax import tree_util
# from meent.on_jax.optimizer.optimizer import OptimizerJax
#
# tree_util.register_pytree_node(OptimizerJax,
#                                OptimizerJax._tree_flatten,
#                                OptimizerJax._tree_unflatten)



import torch


def cell_compression(cell, device=torch.device('cpu'), type_complex=torch.complex128):

    if type_complex == torch.complex128:
        type_float = torch.float64
    else:
        type_float = torch.float32

    # find discontinuities in x
    step_y, step_x = 1. / torch.tensor(cell.shape, device=device, dtype=type_float)
    x = []
    y = []
    cell_x = []
    cell_xy = []

    cell_next = torch.roll(cell, -1, dims=1)

    for col in torch.arange(cell.shape[1]):
        if not (cell[:, col] == cell_next[:, col]).all() or (col == cell.shape[1] - 1):
            x.append(step_x * (col + 1))
            cell_x.append(cell[:, col].reshape((1, -1)))

    cell_x = torch.cat(cell_x, dim=0).T
    cell_x_next = torch.roll(cell_x, -1, dims=0)

    for row in range(cell_x.shape[0]):
        if not (cell_x[row, :] == cell_x_next[row, :]).all() or (row == cell_x.shape[0] - 1):
            y.append(step_y * (row + 1))
            cell_xy.append(cell_x[row, :].reshape((1, -1)))

    x = torch.tensor(x, device=device).reshape((-1, 1)).type(type_complex)
    y = torch.tensor(y, device=device).reshape((-1, 1)).type(type_complex)
    cell_comp = torch.cat(cell_xy, dim=0)

    return cell_comp, x, y


def fft_piecewise_constant(cell, x, y, fourier_order_x, fourier_order_y, device=torch.device('cpu'),
                           type_complex=torch.complex128):

    period_x, period_y = x[-1], y[-1]

    # X axis
    cell_next_x = torch.roll(cell, -1, dims=1)
    cell_diff_x = cell_next_x - cell
    cell_diff_x = cell_diff_x.type(type_complex)

    cell = cell.type(type_complex)

    modes_x = torch.arange(-2 * fourier_order_x, 2 * fourier_order_x + 1, 1, device=device).type(type_complex)

    f_coeffs_x = cell_diff_x @ torch.exp(-1j * 2 * torch.pi * x @ modes_x[None, :] / period_x).type(type_complex)
    c = f_coeffs_x.shape[1] // 2

    x_next = torch.vstack((torch.roll(x, -1, dims=0)[:-1], torch.tensor([period_x], device=device))) - x

    f_coeffs_x[:, c] = (cell @ torch.vstack((x[0], x_next[:-1]))).flatten() / period_x
    mask = torch.ones(f_coeffs_x.shape[1], device=device).type(torch.bool)
    mask[c] = False
    f_coeffs_x[:, mask] /= (1j * 2 * torch.pi * modes_x[mask])

    # Y axis
    f_coeffs_x_next_y = torch.roll(f_coeffs_x, -1, dims=0)
    f_coeffs_x_diff_y = f_coeffs_x_next_y - f_coeffs_x

    modes_y = torch.arange(-2 * fourier_order_y, 2 * fourier_order_y + 1, 1, device=device).type(type_complex)

    f_coeffs_xy = f_coeffs_x_diff_y.T @ torch.exp(-1j * 2 * torch.pi * y @ modes_y[None, :] / period_y)
    c = f_coeffs_xy.shape[1] // 2

    y_next = torch.vstack((torch.roll(y, -1, dims=0)[:-1], torch.tensor([period_y], device=device))) - y

    f_coeffs_xy[:, c] = f_coeffs_x.T @ torch.vstack((y[0], y_next[:-1])).flatten() / period_y

    if c:
        mask = torch.ones(f_coeffs_xy.shape[1], device=device).type(torch.bool)
        mask[c] = False
        f_coeffs_xy[:, mask] /= (1j * 2 * torch.pi * modes_y[mask])

    return f_coeffs_xy.T


def to_conv_mat_vector(ucell_info_list, fourier_order_x, fourier_order_y, device=torch.device('cpu'),
                       type_complex=torch.complex128):
    ff_x = 2 * fourier_order_x + 1
    ff_y = 2 * fourier_order_y + 1

    e_conv_all = torch.zeros((len(ucell_info_list), ff_x * ff_y, ff_x * ff_y), device=device).type(type_complex)
    o_e_conv_all = torch.zeros((len(ucell_info_list), ff_x * ff_y, ff_x * ff_y), device=device).type(type_complex)

    # 2D
    for i, (cell, x_list, y_list) in enumerate(ucell_info_list):

        f_coeffs = fft_piecewise_constant(cell**2, x_list, y_list,
                                          fourier_order_x, fourier_order_y, device=device, type_complex=type_complex)
        o_f_coeffs = fft_piecewise_constant(1 / cell**2, x_list, y_list,
                                            fourier_order_x, fourier_order_y, device=device, type_complex=type_complex)

        center = torch.tensor(f_coeffs.shape, device=device) // 2

        conv_idx_y = torch.arange(-ff_y + 1, ff_y, 1, device=device)
        conv_idx_y = circulant(conv_idx_y, device=device)
        conv_i = conv_idx_y.repeat_interleave(ff_x, dim=1)
        conv_i = conv_i.repeat_interleave(ff_x, dim=0)

        conv_idx_x = torch.arange(-ff_x + 1, ff_x, 1, device=device)
        conv_idx_x = circulant(conv_idx_x, device=device)
        conv_j = conv_idx_x.repeat(ff_y, ff_y)

        e_conv = f_coeffs[center[0] + conv_i, center[1] + conv_j]
        o_e_conv = o_f_coeffs[center[0] + conv_i, center[1] + conv_j]

        e_conv_all[i] = e_conv
        o_e_conv_all[i] = o_e_conv
    return e_conv_all, o_e_conv_all


# TODO: Fourier analysis - Analytic
# def fs_rectangle(cx, cy, lx, ly, pmtvy, fourier_order, period):
#     Px, Py = period
#     fourier_x, fourier_y = fourier_order
#     Nx_vector = torch.arange(-2 * fourier_x, 2 * fourier_x + 1, 1).type(torch.complex128)
#     Ny_vector = torch.arange(-2 * fourier_y, 2 * fourier_y + 1, 1).type(torch.complex128)
#
#     # Nx_vector = torch.arange(-fourier_x, fourier_x + 1, 1).type(torch.complex128) * torch.cos(torch.tensor(torch.pi/4))
#     # Ny_vector = torch.arange(-fourier_y, fourier_y + 1, 1).type(torch.complex128) * torch.sin(torch.tensor(torch.pi/4))
#
#     # c = n_index * lx * ly / Px / Py * torch.sinc(torch.pi * ly / Py * Ny_vector.reshape((-1, 1))) @ torch.sinc(torch.pi * lx / Px * Nx_vector.reshape((1, -1)))
#     c = pmtvy * lx * ly / Px / Py * torch.sinc(ly / Py * Ny_vector.reshape((-1, 1))) @ torch.sinc(
#         lx / Px * Nx_vector.reshape((1, -1)))
#
#     # c = c * (torch.exp(-1j * 2 * torch.pi * cy / Py * Ny_vector.reshape((-1, 1))) @ torch.exp(-1j * 2 * torch.pi * cx / Px * Nx_vector.reshape((1, -1))))
#     c = c * (torch.exp(-1j * 2 * torch.pi * cy / Py * Ny_vector.reshape((-1, 1))) @ torch.exp(
#         -1j * 2 * torch.pi * cx / Px * Nx_vector.reshape((1, -1))))
#
#     return c.T
#
#
# def to_conv_mat_continuous_vector_sinc(ucell_info_list, fourier_order_x, fourier_order_y, device=torch.device('cpu'),
#                                   type_complex=torch.complex128):
#     ff_x = 2 * fourier_order_x + 1
#     ff_y = 2 * fourier_order_y + 1
#
#     e_conv_all = torch.zeros((len(ucell_info_list), ff_x * ff_y, ff_x * ff_y)).type(type_complex)
#     o_e_conv_all = torch.zeros((len(ucell_info_list), ff_x * ff_y, ff_x * ff_y)).type(type_complex)
#
#     # 2D
#     for i, ucell_info in enumerate(ucell_info_list):
#         ucell_layer, x_list, y_list = ucell_info
#         ucell_layer = ucell_layer ** 2
#         # ucell_layer = torch.tensor(ucell_layer, dtype=type_complex) if type(ucell_layer) != torch.Tensor else ucell_layer
#         # x_list = torch.tensor(x_list, dtype=type_complex) if type(x_list) != torch.Tensor else x_list
#         # y_list = torch.tensor(y_list, dtype=type_complex) if type(y_list) != torch.Tensor else y_list
#
#         f_coeffs = fft_piecewise_constant_vector(ucell_layer, x_list, y_list,
#                                                  fourier_order_x, fourier_order_y, type_complex=type_complex)
#         o_f_coeffs = fft_piecewise_constant_vector(1/ucell_layer, x_list, y_list,
#                                                  fourier_order_x, fourier_order_y, type_complex=type_complex)
#
#         import seaborn as sns
#         import matplotlib.pyplot as plt
#         sns.heatmap(f_coeffs.detach().real)
#         plt.show()
#         sns.heatmap(o_f_coeffs.detach().real)
#         plt.show()
#
#         sns.heatmap(f_coeffs.detach().imag)
#         plt.show()
#         sns.heatmap(o_f_coeffs.detach().imag)
#         plt.show()
#
#         # f_coeffs = fs_rectangle()
#
#
#         center = torch.tensor(f_coeffs.shape, device=device) // 2
#         # center = torch.div(torch.tensor(f_coeffs.shape, device=device), 2, rounding_mode='trunc')
#         # center = torch.tensor(center, device=device)
#
#         conv_idx_y = torch.arange(-ff_y + 1, ff_y, 1, device=device)
#         conv_idx_y = circulant(conv_idx_y, device=device)
#         conv_i = conv_idx_y.repeat_interleave(ff_x, dim=1)
#         conv_i = conv_i.repeat_interleave(ff_x, dim=0)
#
#         conv_idx_x = torch.arange(-ff_x + 1, ff_x, 1, device=device)
#         conv_idx_x = circulant(conv_idx_x, device=device)
#         conv_j = conv_idx_x.repeat(ff_y, ff_y)
#
#         e_conv = f_coeffs[center[0] + conv_i, center[1] + conv_j]
#         o_e_conv = o_f_coeffs[center[0] + conv_i, center[1] + conv_j]
#
#         e_conv_all[i] = e_conv
#         o_e_conv_all[i] = o_e_conv
#     return e_conv_all, o_e_conv_all


def to_conv_mat_raster_continuous(ucell, fourier_order_x, fourier_order_y, device=torch.device('cpu'),
                                  type_complex=torch.complex128):
    ucell_pmt = ucell ** 2

    if ucell_pmt.shape[1] == 1:  # 1D
        ff = 2 * fourier_order_x + 1

        e_conv_all = torch.zeros((ucell_pmt.shape[0], ff, ff), device=device).type(type_complex)
        o_e_conv_all = torch.zeros((ucell_pmt.shape[0], ff, ff), device=device).type(type_complex)

        for i, layer in enumerate(ucell_pmt):

            cell, x, y = cell_compression(layer, device=device, type_complex=type_complex)

            f_coeffs = fft_piecewise_constant(cell, x, y, fourier_order_x, fourier_order_y, device=device, type_complex=type_complex)
            o_f_coeffs = fft_piecewise_constant(1 / cell, x, y, fourier_order_x, fourier_order_y, device=device, type_complex=type_complex)

            center = torch.tensor(f_coeffs.shape, device=device) // 2
            conv_idx = torch.arange(-ff + 1, ff, 1, device=device)
            conv_idx = circulant(conv_idx, device=device)
            e_conv = f_coeffs[center[0], center[1] + conv_idx]
            o_e_conv = o_f_coeffs[center[0], center[1] + conv_idx]
            e_conv_all[i] = e_conv
            o_e_conv_all[i] = o_e_conv

    else:  # 2D
        ff_x = 2 * fourier_order_x + 1
        ff_y = 2 * fourier_order_y + 1

        e_conv_all = torch.zeros((ucell_pmt.shape[0], ff_x * ff_y,  ff_x * ff_y), device=device).type(type_complex)
        o_e_conv_all = torch.zeros((ucell_pmt.shape[0], ff_x * ff_y,  ff_x * ff_y), device=device).type(type_complex)

        for i, layer in enumerate(ucell_pmt):

            cell, x, y = cell_compression(layer, device=device, type_complex=type_complex)

            f_coeffs = fft_piecewise_constant(cell, x, y, fourier_order_x, fourier_order_y, device=device,
                                              type_complex=type_complex)
            o_f_coeffs = fft_piecewise_constant(1 / cell, x, y, fourier_order_x, fourier_order_y, device=device,
                                                type_complex=type_complex)

            center = torch.tensor(f_coeffs.shape, device=device) // 2

            conv_idx_y = torch.arange(-ff_y + 1, ff_y, 1, device=device)
            conv_idx_y = circulant(conv_idx_y, device=device)
            conv_i = conv_idx_y.repeat_interleave(ff_x, dim=1)
            conv_i = conv_i.repeat_interleave(ff_x, dim=0)

            conv_idx_x = torch.arange(-ff_x + 1, ff_x, 1, device=device)
            conv_idx_x = circulant(conv_idx_x, device=device)
            conv_j = conv_idx_x.repeat(ff_y, ff_y)

            e_conv = f_coeffs[center[0] + conv_i, center[1] + conv_j]
            o_e_conv = o_f_coeffs[center[0] + conv_i, center[1] + conv_j]
            e_conv_all[i] = e_conv
            o_e_conv_all[i] = o_e_conv
    return e_conv_all, o_e_conv_all


def to_conv_mat_raster_discrete(ucell, fourier_order_x, fourier_order_y, device=torch.device('cpu'), type_complex=torch.complex128,
                                improve_dft=True):
    ucell_pmt = ucell ** 2

    if ucell_pmt.shape[1] == 1:  # 1D
        ff = 2 * fourier_order_x + 1
        e_conv_all = torch.zeros((ucell_pmt.shape[0], ff, ff), device=device).type(type_complex)
        o_e_conv_all = torch.zeros((ucell_pmt.shape[0], ff, ff), device=device).type(type_complex)
        if improve_dft:
            minimum_pattern_size = 2 * ff * ucell_pmt.shape[2]
        else:
            minimum_pattern_size = 2 * ff

        for i, layer in enumerate(ucell_pmt):
            n = minimum_pattern_size // layer.shape[1]
            layer = layer.repeat_interleave(n + 1, axis=1)
            f_coeffs = torch.fft.fftshift(torch.fft.fft(layer / layer.numel()).type(type_complex))
            o_f_coeffs = torch.fft.fftshift(torch.fft.fft(1/layer / layer.numel()).type(type_complex))
            center = torch.tensor(f_coeffs.shape, device=device) // 2
            # center = torch.div(torch.tensor(f_coeffs.shape, device=device), 2, rounding_mode='trunc')
            # center = torch.tensor(center, device=device)
            conv_idx = torch.arange(-ff + 1, ff, 1, device=device)
            conv_idx = circulant(conv_idx, device=device)
            e_conv = f_coeffs[center[0], center[1] + conv_idx]
            o_e_conv = o_f_coeffs[center[0], center[1] + conv_idx]
            e_conv_all[i] = e_conv
            o_e_conv_all[i] = o_e_conv

    else:  # 2D
        ff_x = 2 * fourier_order_x + 1
        ff_y = 2 * fourier_order_y + 1

        e_conv_all = torch.zeros((ucell_pmt.shape[0], ff_x * ff_y, ff_x * ff_y), device=device).type(type_complex)
        o_e_conv_all = torch.zeros((ucell_pmt.shape[0], ff_x * ff_y, ff_x * ff_y), device=device).type(type_complex)

        if improve_dft:
            minimum_pattern_size_y = 2 * ff_y * ucell_pmt.shape[1]
            minimum_pattern_size_x = 2 * ff_x * ucell_pmt.shape[2]
        else:
            minimum_pattern_size_y = 2 * ff_y
            minimum_pattern_size_x = 2 * ff_x
        # e.g., 8 bytes * (40*500) * (40*500) / 1E6 = 3200 MB = 3.2 GB

        for i, layer in enumerate(ucell_pmt):
            if layer.shape[0] < minimum_pattern_size_y:
                # n = torch.div(minimum_pattern_size_y, layer.shape[0], rounding_mode='trunc')
                n = minimum_pattern_size_y // layer.shape[0]
                n = torch.tensor(n, device=device)

                layer = layer.repeat_interleave(n + 1, axis=0)
            if layer.shape[1] < minimum_pattern_size_x:
                # n = torch.div(minimum_pattern_size_x, layer.shape[1], rounding_mode='trunc')
                n = minimum_pattern_size_x // layer.shape[1]
                n = torch.tensor(n, device=device)
                layer = layer.repeat_interleave(n + 1, axis=1)

            f_coeffs = torch.fft.fftshift(torch.fft.fft2(layer / layer.numel()))
            o_f_coeffs = torch.fft.fftshift(torch.fft.fft2(1/layer / layer.numel()))
            center = torch.tensor(f_coeffs.shape, device=device) // 2

            conv_idx_y = torch.arange(-ff_y + 1, ff_y, 1, device=device)
            conv_idx_y = circulant(conv_idx_y, device=device)
            conv_i = conv_idx_y.repeat_interleave(ff_x, dim=1)
            conv_i = conv_i.repeat_interleave(ff_x, dim=0)

            conv_idx_x = torch.arange(-ff_x + 1, ff_x, 1, device=device)
            conv_idx_x = circulant(conv_idx_x, device=device)
            conv_j = conv_idx_x.repeat(ff_y, ff_y)

            e_conv = f_coeffs[center[0] + conv_i, center[1] + conv_j]
            o_e_conv = o_f_coeffs[center[0] + conv_i, center[1] + conv_j]
            e_conv_all[i] = e_conv
            o_e_conv_all[i] = o_e_conv
    return e_conv_all, o_e_conv_all


def circulant(c, device=torch.device('cpu')):

    center = c.shape[0] // 2
    circ = torch.zeros((center + 1, center + 1), device=device).type(torch.long)

    for r in range(center+1):
        idx = torch.arange(r, r - center - 1, -1, device=device)

        assign_value = c[center - idx]
        circ[r] = assign_value

    return circ



import torch
import numpy as np

from .modeler.modeling import ModelingTorch
from .emsolver.rcwa import RCWATorch
from .optimizer.optimizer import OptimizerTorch


class MeeTorch(ModelingTorch, RCWATorch, OptimizerTorch):

    def __init__(self, device=0, type_complex=0, *args, **kwargs):

        # device
        if device in (0, 'cpu'):
            self._device = torch.device('cpu')
        elif device in (1, 'gpu', 'cuda'):
            self._device = torch.device('cuda')
        elif type(device) is torch.device:
            self._device = device
        else:
            raise ValueError('device')

        # type_complex
        if type_complex in (0, torch.complex128, np.complex128):
            self._type_complex = torch.complex128
        elif type_complex in (1, torch.complex64, np.complex64):
            self._type_complex = torch.complex64
        else:
            raise ValueError('Torch type_complex')

        self._type_float = torch.float64 if self._type_complex is not torch.complex64 else torch.float32
        self._type_int = torch.int64 if self._type_complex is not torch.complex64 else torch.int32
        # self.perturbation = perturbation

        self.device = device
        self.type_complex = type_complex

        ModelingTorch.__init__(self, device=device, type_complex=type_complex, *args, **kwargs)
        RCWATorch.__init__(self, device=device, type_complex=type_complex, *args, **kwargs)
        OptimizerTorch.__init__(self, device=device, type_complex=type_complex, *args, **kwargs)



import time
import torch

import numpy as np

from ._base import _BaseRCWA
from .convolution_matrix import to_conv_mat_raster_discrete, to_conv_mat_raster_continuous, to_conv_mat_vector
from .field_distribution import field_dist_1d_vanilla, field_dist_2d_vanilla, field_plot, field_dist_1d_conical_vanilla, \
    field_dist_1d_vectorized_ji, field_dist_1d_vectorized_kji, field_dist_1d_conical_vectorized_ji, \
    field_dist_1d_conical_vectorized_kji, field_dist_2d_vectorized_ji, field_dist_2d_vectorized_kji


class RCWATorch(_BaseRCWA):
    def __init__(self,
                 n_I=1.,
                 n_II=1.,
                 theta=0.,
                 phi=0.,
                 period=(100., 100.),
                 wavelength=900.,
                 ucell=None,
                 ucell_info_list=None,
                 thickness=(0., ),
                 backend=2,
                 grating_type=0,
                 pol=0.,
                 fourier_order=(2, 0),
                 ucell_materials=None,
                 algo='TMM',
                 perturbation=1E-20,
                 device='cpu',
                 type_complex=torch.complex128,
                 fft_type=0,
                 improve_dft=True,
                 **kwargs,
                 ):

        super().__init__(grating_type=grating_type, n_I=n_I, n_II=n_II, theta=theta, phi=phi, pol=pol,
                         fourier_order=fourier_order, period=period, wavelength=wavelength,
                         thickness=thickness, algo=algo, perturbation=perturbation,
                         device=device, type_complex=type_complex)

        self.ucell = ucell
        self.ucell_materials = ucell_materials
        self.ucell_info_list = ucell_info_list

        self.backend = backend
        self.fft_type = fft_type
        self.improve_dft = improve_dft

        self.layer_info_list = []

    @property
    def ucell(self):
        return self._ucell

    @ucell.setter
    def ucell(self, ucell):
        if type(ucell) is torch.Tensor:
            if ucell.dtype in (torch.complex128, torch.complex64):
                dtype = self.type_complex
            elif ucell.dtype in (torch.float64, torch.float32, torch.int64, torch.int32):
                dtype = self.type_float
            else:
                raise ValueError
            self._ucell = ucell.to(device=self.device, dtype=dtype)
        elif isinstance(ucell, np.ndarray):
            if ucell.dtype in (np.int64, np.float64, np.int32, np.float32):
                dtype = self.type_float
            elif ucell.dtype in (np.complex128, np.complex64):
                dtype = self.type_complex
            else:
                raise ValueError
            self._ucell = torch.tensor(ucell, device=self.device, dtype=dtype)
        elif ucell is None:
            self._ucell = ucell
        else:
            raise ValueError

    def _solve(self, wavelength, e_conv_all, o_e_conv_all):
        self.kx_vector = self.get_kx_vector(wavelength)

        if self.grating_type == 0:
            de_ri, de_ti, rayleigh_r, rayleigh_t, layer_info_list, T1 = self.solve_1d(wavelength, e_conv_all, o_e_conv_all)
        elif self.grating_type == 1:
            de_ri, de_ti, rayleigh_r, rayleigh_t, layer_info_list, T1 = self.solve_1d_conical(wavelength, e_conv_all, o_e_conv_all)
        elif self.grating_type == 2:
            de_ri, de_ti, rayleigh_r, rayleigh_t, layer_info_list, T1 = self.solve_2d(wavelength, e_conv_all, o_e_conv_all)
        else:
            raise ValueError

        return de_ri, de_ti, rayleigh_r, rayleigh_t, layer_info_list, T1, self.kx_vector

    def solve(self, wavelength, e_conv_all, o_e_conv_all):
        de_ri, de_ti, rayleigh_r, rayleigh_t, layer_info_list, T1, kx_vector = self._solve(wavelength, e_conv_all, o_e_conv_all)

        self.rayleigh_r = rayleigh_r
        self.rayleigh_t = rayleigh_t
        self.layer_info_list = layer_info_list
        self.T1 = T1
        self.kx_vector = kx_vector

        return de_ri, de_ti

    def conv_solve(self, **kwargs):
        [setattr(self, k, v) for k, v in kwargs.items()]  # needed for optimization

        if self.fft_type == 0:
            E_conv_all, o_E_conv_all = to_conv_mat_raster_discrete(self.ucell, self.fourier_order[0], self.fourier_order[1],
                                                                   device=self.device, type_complex=self.type_complex,
                                                                   improve_dft=self.improve_dft)
        elif self.fft_type == 1:
            E_conv_all, o_E_conv_all = to_conv_mat_raster_continuous(self.ucell, self.fourier_order[0], self.fourier_order[1],
                                                                     device=self.device, type_complex=self.type_complex)
        elif self.fft_type == 2:
            E_conv_all, o_E_conv_all = to_conv_mat_vector(self.ucell_info_list, self.fourier_order[0],
                                                          self.fourier_order[1],
                                                          type_complex=self.type_complex)

        else:
            raise ValueError

        de_ri, de_ti, rayleigh_r, rayleigh_t, layer_info_list, T1, kx_vector = self._solve(self.wavelength, E_conv_all, o_E_conv_all)

        self.layer_info_list = layer_info_list
        self.T1 = T1
        self.kx_vector = kx_vector

        return de_ri, de_ti

    def calculate_field(self, res_x=20, res_y=20, res_z=20, field_algo=2):
        if self.grating_type == 0:
            res_y = 1
            if field_algo == 0:
                field_cell = field_dist_1d_vanilla(self.wavelength, self.kx_vector,
                                                   self.T1, self.layer_info_list, self.period, self.pol,
                                                   res_x=res_x, res_y=res_y, res_z=res_z,
                                                   device=self.device, type_complex=self.type_complex)
            elif field_algo == 1:
                field_cell = field_dist_1d_vectorized_ji(self.wavelength, self.kx_vector, self.T1, self.layer_info_list,
                                                         self.period, self.pol, res_x=res_x, res_y=res_y, res_z=res_z,
                                                         device=self.device, type_complex=self.type_complex,
                                                         type_float=self.type_float)
            elif field_algo == 2:
                field_cell = field_dist_1d_vectorized_kji(self.wavelength, self.kx_vector, self.T1,
                                                          self.layer_info_list, self.period, self.pol,
                                                          res_x=res_x, res_y=res_y, res_z=res_z,
                                                          device=self.device, type_complex=self.type_complex,
                                                          type_float=self.type_float)
            else:
                raise ValueError
        elif self.grating_type == 1:
            res_y = 1
            if field_algo == 0:
                field_cell = field_dist_1d_conical_vanilla(self.wavelength, self.kx_vector, self.n_I, self.theta,
                                                           self.phi, self.T1, self.layer_info_list, self.period,
                                                           res_x=res_x, res_y=res_y, res_z=res_z, device=self.device,
                                                           type_complex=self.type_complex)
            elif field_algo == 1:
                field_cell = field_dist_1d_conical_vectorized_ji(self.wavelength, self.kx_vector, self.n_I, self.theta,
                                                                 self.phi, self.T1, self.layer_info_list, self.period,
                                                                 res_x=res_x, res_y=res_y, res_z=res_z,
                                                                 device=self.device,
                                                                 type_complex=self.type_complex,
                                                                 type_float=self.type_float)
            elif field_algo == 2:
                field_cell = field_dist_1d_conical_vectorized_kji(self.wavelength, self.kx_vector, self.n_I, self.theta,
                                                                  self.phi, self.T1, self.layer_info_list, self.period,
                                                                  res_x=res_x, res_y=res_y, res_z=res_z,
                                                                  device=self.device,
                                                                  type_complex=self.type_complex,
                                                                  type_float=self.type_float)
            else:
                raise ValueError

        elif self.grating_type == 2:
            if field_algo == 0:
                field_cell = field_dist_2d_vanilla(self.wavelength, self.kx_vector, self.n_I, self.theta, self.phi,
                                                   self.fourier_order[0], self.fourier_order[1], self.T1,
                                                   self.layer_info_list, self.period,
                                                   res_x=res_x, res_y=res_y, res_z=res_z, device=self.device,
                                                   type_complex=self.type_complex, type_float=self.type_float)
            elif field_algo == 1:
                field_cell = field_dist_2d_vectorized_ji(self.wavelength, self.kx_vector, self.n_I, self.theta,
                                                         self.phi, self.fourier_order[0], self.fourier_order[1],
                                                         self.T1, self.layer_info_list,
                                                         self.period, res_x=res_x, res_y=res_y, res_z=res_z,
                                                         device=self.device,
                                                         type_complex=self.type_complex, type_float=self.type_float)
            elif field_algo == 2:
                field_cell = field_dist_2d_vectorized_kji(self.wavelength, self.kx_vector, self.n_I, self.theta,
                                                          self.phi, self.fourier_order[0], self.fourier_order[1],
                                                          self.T1, self.layer_info_list,
                                                          self.period, res_x=res_x, res_y=res_y, res_z=res_z,
                                                          device=self.device,
                                                          type_complex=self.type_complex, type_float=self.type_float)
            else:
                raise ValueError
        else:
            raise ValueError

        return field_cell

    def conv_solve_field(self, res_x=20, res_y=20, res_z=20, field_algo=2):
        de_ri, de_ti = self.conv_solve()
        field_cell = self.calculate_field(res_x, res_y, res_z, field_algo=field_algo)
        return de_ri, de_ti, field_cell

    def field_plot(self, field_cell):
        field_plot(field_cell, self.pol)

    def calculate_field_all(self, res_x=20, res_y=20, res_z=20):
        t0 = time.time()
        field_cell0 = self.calculate_field(res_x=res_x, res_y=res_y, res_z=res_z, field_algo=0)
        print('no vector', time.time() - t0)
        t0 = time.time()
        field_cell1 = self.calculate_field(res_x=res_x, res_y=res_y, res_z=res_z, field_algo=1)
        print('ji vector', time.time() - t0)
        t0 = time.time()
        field_cell2 = self.calculate_field(res_x=res_x, res_y=res_y, res_z=res_z, field_algo=2)
        print('kji vector', time.time() - t0)

        print('gap(1-0): ', torch.linalg.norm(field_cell1 - field_cell0))
        print('gap(2-1): ', torch.linalg.norm(field_cell2 - field_cell1))
        print('gap(0-2): ', torch.linalg.norm(field_cell0 - field_cell2))

        return field_cell0, field_cell1, field_cell2



import jax
import jax.numpy as jnp

from functools import partial

# @partial(jax.jit, static_argnums=(5, 6, 7, 8, 9))
def field_dist_1d_vectorized_ji(wavelength, kx_vector, T1, layer_info_list, period,
                                pol, res_x=20, res_y=20, res_z=20, type_complex=jnp.complex128, type_float=jnp.float64):

    k0 = 2 * jnp.pi / wavelength
    Kx = jnp.diag(kx_vector / k0)

    field_cell = jnp.zeros((res_z * len(layer_info_list), res_y, res_x, 3), dtype=type_complex)

    T_layer = T1

    # From the first layer
    for idx_layer, (E_conv_i, q, W, X, a_i, b, d) in enumerate(layer_info_list[::-1]):

        c1 = T_layer[:, None]
        c2 = b @ a_i @ X @ T_layer[:, None]

        Q = jnp.diag(q)

        if pol == 0:
            V = W @ Q
            EKx = None

        else:
            V = E_conv_i @ W @ Q
            EKx = E_conv_i @ Kx

        for k in range(res_z):
            z = k / res_z * d

            if pol == 0:
                Sy = W @ (diag_exp(-k0 * Q * z) @ c1 + diag_exp(k0 * Q * (z - d)) @ c2)
                Ux = V @ (-diag_exp(-k0 * Q * z) @ c1 + diag_exp(k0 * Q * (z - d)) @ c2)
                C = Kx @ Sy

                x_1d = jnp.arange(res_x, dtype=type_float).reshape((1, -1, 1))
                x_1d = -1j * x_1d * period[0] / res_x
                x_2d = jnp.tile(x_1d, (res_y, 1, 1))
                x_2d = x_2d * kx_vector
                x_2d = x_2d.reshape((res_y, res_x, 1, len(kx_vector)))

                exp_K = jnp.exp(x_2d)
                exp_K = exp_K.reshape((res_y, res_x, -1))

                Ey = exp_K @ Sy
                Hx = -1j * exp_K @ Ux
                Hz = -1j * exp_K @ C

                val = jnp.concatenate((Ey, Hx, Hz), axis=-1)

            else:
                Uy = W @ (diag_exp(-k0 * Q * z) @ c1 + diag_exp(k0 * Q * (z - d)) @ c2)
                Sx = V @ (-diag_exp(-k0 * Q * z) @ c1 + diag_exp(k0 * Q * (z - d)) @ c2)

                C = EKx @ Uy  # there is a better option for convergence

                x_1d = jnp.arange(res_x, dtype=type_float).reshape((1, -1, 1))
                x_1d = -1j * x_1d * period[0] / res_x
                x_2d = jnp.tile(x_1d, (res_y, 1, 1))
                x_2d = x_2d * kx_vector
                x_2d = x_2d.reshape((res_y, res_x, 1, len(kx_vector)))

                exp_K = jnp.exp(x_2d)
                exp_K = exp_K.reshape((res_y, res_x, -1))

                Hy = exp_K @ Uy
                Ex = 1j * exp_K @ Sx
                Ez = -1j * exp_K @ C

                val = jnp.concatenate((Hy, Ex, Ez), axis=-1)

            field_cell = field_cell.at[res_z * idx_layer + k].set(val)

        T_layer = a_i @ X @ T_layer

    return field_cell


# @partial(jax.jit, static_argnums=(8, 9, 10, 11))
def field_dist_1d_conical_vectorized_ji(wavelength, kx_vector, n_I, theta, phi, T1, layer_info_list, period,
                                        res_x=20, res_y=20, res_z=20, type_complex=jnp.complex128, type_float=jnp.float64):

    k0 = 2 * jnp.pi / wavelength
    ky = k0 * n_I * jnp.sin(theta) * jnp.sin(phi)
    Kx = jnp.diag(kx_vector / k0)

    field_cell = jnp.zeros((res_z * len(layer_info_list), res_y, res_x, 6), dtype=type_complex)

    T_layer = T1

    big_I = jnp.eye((len(T1))).astype(type_complex)

    # From the first layer
    for idx_layer, [E_conv_i, q_1, q_2, W_1, W_2, V_11, V_12, V_21, V_22, big_X, big_A_i, big_B, d] \
            in enumerate(layer_info_list[::-1]):

        c = jnp.block([[big_I], [big_B @ big_A_i @ big_X]]) @ T_layer

        ff = len(c) // 4

        c1_plus = c[0 * ff:1 * ff]
        c2_plus = c[1 * ff:2 * ff]
        c1_minus = c[2 * ff:3 * ff]
        c2_minus = c[3 * ff:4 * ff]

        big_Q1 = jnp.diag(q_1)
        big_Q2 = jnp.diag(q_2)
        for k in range(res_z):
            # Sx, Sy, Ux, Uy, Sz, Uz = z_loop_1d_conical(k, c, k0, Kx, ky, res_z, E_conv_i, q_1, q_2, W_1, W_2, V_11, V_12, V_21, V_22, d)

            z = k / res_z * d

            Sx = W_2 @ (diag_exp(-k0 * big_Q2 * z) @ c2_plus + diag_exp(k0 * big_Q2 * (z - d)) @ c2_minus)

            Sy = V_11 @ (diag_exp(-k0 * big_Q1 * z) @ c1_plus + diag_exp(k0 * big_Q1 * (z - d)) @ c1_minus) \
                 + V_12 @ (diag_exp(-k0 * big_Q2 * z) @ c2_plus + diag_exp(k0 * big_Q2 * (z - d)) @ c2_minus)

            Ux = W_1 @ (-diag_exp(-k0 * big_Q1 * z) @ c1_plus + diag_exp(k0 * big_Q1 * (z - d)) @ c1_minus)

            Uy = V_21 @ (-diag_exp(-k0 * big_Q1 * z) @ c1_plus + diag_exp(k0 * big_Q1 * (z - d)) @ c1_minus) \
                 + V_22 @ (-diag_exp(-k0 * big_Q2 * z) @ c2_plus + diag_exp(k0 * big_Q2 * (z - d)) @ c2_minus)

            Sz = -1j * E_conv_i @ (Kx @ Uy - ky * Ux)

            Uz = -1j * (Kx @ Sy - ky * Sx)

            x_1d = jnp.arange(res_x, dtype=type_float).reshape((1, -1, 1))
            x_1d = -1j * x_1d * period[0] / res_x
            x_2d = jnp.tile(x_1d, (res_y, 1, 1))
            x_2d = x_2d * kx_vector
            x_2d = x_2d.reshape((res_y, res_x, 1, len(kx_vector)))

            exp_K = jnp.exp(x_2d)
            exp_K = exp_K.reshape((res_y, res_x, -1))

            Ex = exp_K @ Sx
            Ey = exp_K @ Sy
            Ez = exp_K @ Sz

            Hx = -1j * exp_K @ Ux
            Hy = -1j * exp_K @ Uy
            Hz = -1j * exp_K @ Uz

            val = jnp.concatenate((Ex, Ey, Ez, Hx, Hy, Hz), axis=-1)

            field_cell = field_cell.at[res_z * idx_layer + k].set(val)

        T_layer = big_A_i @ big_X @ T_layer

    return field_cell


# @partial(jax.jit, static_argnums=(5, 6, 10, 11, 12, 13))
def field_dist_2d_vectorized_ji(wavelength, kx_vector, n_I, theta, phi, fourier_order_x, fourier_order_y, T1, layer_info_list, period,
                                res_x=20, res_y=20, res_z=20, type_complex=jnp.complex128, type_float=jnp.float64):

    k0 = 2 * jnp.pi / wavelength

    fourier_indices_y = jnp.arange(-fourier_order_y, fourier_order_y + 1, dtype=type_float)
    ff_x = fourier_order_x * 2 + 1
    ff_y = fourier_order_y * 2 + 1
    ky_vector = k0 * (n_I * jnp.sin(theta) * jnp.sin(phi) + fourier_indices_y * (
            wavelength / period[1])).astype(type_complex)

    Kx = jnp.diag(jnp.tile(kx_vector, ff_y).flatten()) / k0
    Ky = jnp.diag(jnp.tile(ky_vector.reshape((-1, 1)), ff_x).flatten()) / k0

    field_cell = jnp.zeros((res_z * len(layer_info_list), res_y, res_x, 6), dtype=type_complex)

    T_layer = T1

    big_I = jnp.eye((len(T1))).astype(type_complex)

    # From the first layer
    for idx_layer, (E_conv_i, q, W_11, W_12, W_21, W_22, V_11, V_12, V_21, V_22, big_X, big_A_i, big_B, d)\
            in enumerate(layer_info_list[::-1]):

        c = jnp.block([[big_I], [big_B @ big_A_i @ big_X]]) @ T_layer

        ff = len(c) // 4

        c1_plus = c[0 * ff:1 * ff]
        c2_plus = c[1 * ff:2 * ff]
        c1_minus = c[2 * ff:3 * ff]
        c2_minus = c[3 * ff:4 * ff]

        q1 = q[:len(q) // 2]
        q2 = q[len(q) // 2:]
        big_Q1 = jnp.diag(q1)
        big_Q2 = jnp.diag(q2)
        for k in range(res_z):
            z = k / res_z * d
            Sx = W_11 @ (diag_exp(-k0 * big_Q1 * z) @ c1_plus + diag_exp(k0 * big_Q1 * (z - d)) @ c1_minus) \
                 + W_12 @ (diag_exp(-k0 * big_Q2 * z) @ c2_plus + diag_exp(k0 * big_Q2 * (z - d)) @ c2_minus)

            Sy = W_21 @ (diag_exp(-k0 * big_Q1 * z) @ c1_plus + diag_exp(k0 * big_Q1 * (z - d)) @ c1_minus) \
                 + W_22 @ (diag_exp(-k0 * big_Q2 * z) @ c2_plus + diag_exp(k0 * big_Q2 * (z - d)) @ c2_minus)

            Ux = V_11 @ (-diag_exp(-k0 * big_Q1 * z) @ c1_plus + diag_exp(k0 * big_Q1 * (z - d)) @ c1_minus) \
                 + V_12 @ (-diag_exp(-k0 * big_Q2 * z) @ c2_plus + diag_exp(k0 * big_Q2 * (z - d)) @ c2_minus)

            Uy = V_21 @ (-diag_exp(-k0 * big_Q1 * z) @ c1_plus + diag_exp(k0 * big_Q1 * (z - d)) @ c1_minus) \
                 + V_22 @ (-diag_exp(-k0 * big_Q2 * z) @ c2_plus + diag_exp(k0 * big_Q2 * (z - d)) @ c2_minus)

            Sz = -1j * E_conv_i @ (Kx @ Uy - Ky @ Ux)

            Uz = -1j * (Kx @ Sy - Ky @ Sx)

            x_1d = jnp.arange(res_x, dtype=type_float).reshape((1, -1, 1))
            y_1d = jnp.arange(res_y, dtype=type_float).reshape((-1, 1, 1))

            x_1d = -1j * x_1d * period[0] / res_x
            y_1d = -1j * y_1d * period[1] / res_y

            x_2d = jnp.tile(x_1d, (res_y, 1, 1))
            y_2d = jnp.tile(y_1d, (1, res_x, 1))

            x_2d = x_2d * kx_vector
            y_2d = y_2d * ky_vector

            x_2d = x_2d.reshape((res_y, res_x, 1, len(kx_vector)))
            y_2d = y_2d.reshape((res_y, res_x, len(ky_vector), 1))

            exp_K = jnp.exp(x_2d) * jnp.exp(y_2d)
            exp_K = exp_K.reshape((res_y, res_x, -1))

            Ex = exp_K @ Sx
            Ey = exp_K @ Sy
            Ez = exp_K @ Sz

            Hx = -1j * exp_K @ Ux
            Hy = -1j * exp_K @ Uy
            Hz = -1j * exp_K @ Uz

            val = jnp.concatenate((Ex, Ey, Ez, Hx, Hy, Hz), axis=-1)

            field_cell = field_cell.at[res_z * idx_layer + k].set(val)

        T_layer = big_A_i @ big_X @ T_layer

    return field_cell


def field_dist_1d_vectorized_kji(wavelength, kx_vector, T1, layer_info_list, period, pol, res_x=20, res_y=20, res_z=20,
                                 type_complex=jnp.complex128, type_float=jnp.float64):

    k0 = 2 * jnp.pi / wavelength
    Kx = jnp.diag(kx_vector / k0)

    field_cell = jnp.zeros((res_z * len(layer_info_list), res_y, res_x, 3), dtype=type_complex)

    T_layer = T1

    # From the first layer
    for idx_layer, (E_conv_i, q, W, X, a_i, b, d) in enumerate(layer_info_list[::-1]):

        c1 = T_layer[:, None]
        c2 = b @ a_i @ X @ T_layer[:, None]

        Q = jnp.diag(q)

        if pol == 0:
            V = W @ Q
            EKx = None

        else:
            V = E_conv_i @ W @ Q
            EKx = E_conv_i @ Kx

        z_1d = jnp.arange(res_z, dtype=type_float).reshape((-1, 1, 1)) / res_z * d

        if pol == 0:
            Sy = W @ (diag_exp_batch(-k0 * Q * z_1d) @ c1 + diag_exp_batch(k0 * Q * (z_1d - d)) @ c2)
            Ux = V @ (-diag_exp_batch(-k0 * Q * z_1d) @ c1 + diag_exp_batch(k0 * Q * (z_1d - d)) @ c2)
            C = Kx @ Sy

            x_1d = jnp.arange(res_x, dtype=type_float).reshape((1, -1, 1))
            x_1d = -1j * x_1d * period[0] / res_x
            x_2d = jnp.tile(x_1d, (res_y, 1, 1))
            x_2d = x_2d * kx_vector
            x_2d = x_2d.reshape((res_y, res_x, 1, len(kx_vector)))

            exp_K = jnp.exp(x_2d)
            exp_K = exp_K.reshape((res_y, res_x, -1))

            Ey = exp_K[:, :, None, :] @ Sy[:, None, None, :, :]
            Hx = -1j * exp_K[:, :, None, :] @ Ux[:, None, None, :, :]
            Hz = -1j * exp_K[:, :, None, :] @ C[:, None, None, :, :]

            val = jnp.concatenate((Ey.squeeze(-1), Hx.squeeze(-1), Hz.squeeze(-1)), axis=-1)

        else:
            Uy = W @ (diag_exp_batch(-k0 * Q * z_1d) @ c1 + diag_exp_batch(k0 * Q * (z_1d - d)) @ c2)
            Sx = V @ (-diag_exp_batch(-k0 * Q * z_1d) @ c1 + diag_exp_batch(k0 * Q * (z_1d - d)) @ c2)

            C = EKx @ Uy  # there is a better option for convergence

            x_1d = jnp.arange(res_x, dtype=type_float).reshape((1, -1, 1))
            x_1d = -1j * x_1d * period[0] / res_x
            x_2d = jnp.tile(x_1d, (res_y, 1, 1))
            x_2d = x_2d * kx_vector
            x_2d = x_2d.reshape((res_y, res_x, 1, len(kx_vector)))

            exp_K = jnp.exp(x_2d)
            exp_K = exp_K.reshape((res_y, res_x, -1))

            Hy = exp_K[:, :, None, :] @ Uy[:, None, None, :, :]
            Ex = 1j * exp_K[:, :, None, :] @ Sx[:, None, None, :, :]
            Ez = -1j * exp_K[:, :, None, :] @ C[:, None, None, :, :]

            val = jnp.concatenate((Hy.squeeze(-1), Ex.squeeze(-1), Ez.squeeze(-1)), axis=-1)

        field_cell = field_cell.at[res_z * idx_layer:res_z * (idx_layer + 1)].set(val)

        T_layer = a_i @ X @ T_layer

    return field_cell


# @partial(jax.jit, static_argnums=(8, 9, 10, 11))
def field_dist_1d_conical_vectorized_kji(wavelength, kx_vector, n_I, theta, phi, T1, layer_info_list, period,
                                         res_x=20, res_y=20, res_z=20,
                                         type_complex=jnp.complex128, type_float=jnp.float64):

    k0 = 2 * jnp.pi / wavelength
    ky = k0 * n_I * jnp.sin(theta) * jnp.sin(phi)
    Kx = jnp.diag(kx_vector / k0)

    field_cell = jnp.zeros((res_z * len(layer_info_list), res_y, res_x, 6), dtype=type_complex)

    T_layer = T1

    big_I = jnp.eye((len(T1))).astype(type_complex)

    # From the first layer
    for idx_layer, [E_conv_i, q_1, q_2, W_1, W_2, V_11, V_12, V_21, V_22, big_X, big_A_i, big_B, d] \
            in enumerate(layer_info_list[::-1]):

        c = jnp.block([[big_I], [big_B @ big_A_i @ big_X]]) @ T_layer

        z_1d = jnp.arange(res_z, dtype=type_float).reshape((-1, 1, 1)) / res_z * d

        ff = len(c) // 4

        c1_plus = c[0 * ff:1 * ff]
        c2_plus = c[1 * ff:2 * ff]
        c1_minus = c[2 * ff:3 * ff]
        c2_minus = c[3 * ff:4 * ff]

        big_Q1 = jnp.diag(q_1)
        big_Q2 = jnp.diag(q_2)

        Sx = W_2 @ (diag_exp_batch(-k0 * big_Q2 * z_1d) @ c2_plus + diag_exp_batch(k0 * big_Q2 * (z_1d - d)) @ c2_minus)

        Sy = V_11 @ (diag_exp_batch(-k0 * big_Q1 * z_1d) @ c1_plus + diag_exp_batch(k0 * big_Q1 * (z_1d - d)) @ c1_minus) \
             + V_12 @ (diag_exp_batch(-k0 * big_Q2 * z_1d) @ c2_plus + diag_exp_batch(k0 * big_Q2 * (z_1d - d)) @ c2_minus)

        Ux = W_1 @ (-diag_exp_batch(-k0 * big_Q1 * z_1d) @ c1_plus + diag_exp_batch(k0 * big_Q1 * (z_1d - d)) @ c1_minus)

        Uy = V_21 @ (-diag_exp_batch(-k0 * big_Q1 * z_1d) @ c1_plus + diag_exp_batch(k0 * big_Q1 * (z_1d - d)) @ c1_minus) \
             + V_22 @ (-diag_exp_batch(-k0 * big_Q2 * z_1d) @ c2_plus + diag_exp_batch(k0 * big_Q2 * (z_1d - d)) @ c2_minus)

        Sz = -1j * E_conv_i @ (Kx @ Uy - ky * Ux)
        Uz = -1j * (Kx @ Sy - ky * Sx)

        x_1d = jnp.arange(res_x, dtype=type_float).reshape((1, -1, 1))
        x_1d = -1j * x_1d * period[0] / res_x
        x_2d = jnp.tile(x_1d, (res_y, 1, 1))
        x_2d = x_2d * kx_vector
        x_2d = x_2d.reshape((res_y, res_x, 1, len(kx_vector)))

        exp_K = jnp.exp(x_2d)
        exp_K = exp_K.reshape((res_y, res_x, -1))

        Ex = exp_K[:, :, None, :] @ Sx[:, None, None, :, :]
        Ey = exp_K[:, :, None, :] @ Sy[:, None, None, :, :]
        Ez = exp_K[:, :, None, :] @ Sz[:, None, None, :, :]

        Hx = -1j * exp_K[:, :, None, :] @ Ux[:, None, None, :, :]
        Hy = -1j * exp_K[:, :, None, :] @ Uy[:, None, None, :, :]
        Hz = -1j * exp_K[:, :, None, :] @ Uz[:, None, None, :, :]

        val = jnp.concatenate(
            (Ex.squeeze(-1), Ey.squeeze(-1), Ez.squeeze(-1), Hx.squeeze(-1), Hy.squeeze(-1), Hz.squeeze(-1)),
            axis=-1)
        field_cell = field_cell.at[res_z * idx_layer:res_z * (idx_layer + 1)].set(val)
        T_layer = big_A_i @ big_X @ T_layer

    return field_cell

# @partial(jax.jit, static_argnums=(5, 6, 10, 11, 12, 13))
def field_dist_2d_vectorized_kji(wavelength, kx_vector, n_I, theta, phi, fourier_order_x, fourier_order_y, T1, layer_info_list, period,
                                 res_x=20, res_y=20, res_z=20, type_complex=jnp.complex128, type_float=jnp.float64):

    k0 = 2 * jnp.pi / wavelength

    fourier_indices_y = jnp.arange(-fourier_order_y, fourier_order_y + 1, dtype=type_float)
    ff_x = fourier_order_x * 2 + 1
    ff_y = fourier_order_y * 2 + 1
    ky_vector = k0 * (n_I * jnp.sin(theta) * jnp.sin(phi) + fourier_indices_y * (
            wavelength / period[1])).astype(type_complex)

    Kx = jnp.diag(jnp.tile(kx_vector, ff_y).flatten()) / k0
    Ky = jnp.diag(jnp.tile(ky_vector.reshape((-1, 1)), ff_x).flatten()) / k0

    field_cell = jnp.zeros((res_z * len(layer_info_list), res_y, res_x, 6), dtype=type_complex)

    T_layer = T1

    big_I = jnp.eye((len(T1))).astype(type_complex)

    # From the first layer
    for idx_layer, (E_conv_i, q, W_11, W_12, W_21, W_22, V_11, V_12, V_21, V_22, big_X, big_A_i, big_B, d)\
            in enumerate(layer_info_list[::-1]):

        c = jnp.block([[big_I], [big_B @ big_A_i @ big_X]]) @ T_layer
        z_1d = jnp.arange(res_z, dtype=type_float).reshape((-1, 1, 1)) / res_z * d

        ff = len(c) // 4

        c1_plus = c[0 * ff:1 * ff]
        c2_plus = c[1 * ff:2 * ff]
        c1_minus = c[2 * ff:3 * ff]
        c2_minus = c[3 * ff:4 * ff]

        q1 = q[:len(q) // 2]
        q2 = q[len(q) // 2:]
        big_Q1 = jnp.diag(q1)
        big_Q2 = jnp.diag(q2)

        Sx = W_11 @ (diag_exp_batch(-k0 * big_Q1 * z_1d) @ c1_plus + diag_exp_batch(k0 * big_Q1 * (z_1d - d)) @ c1_minus) \
              + W_12 @ (diag_exp_batch(-k0 * big_Q2 * z_1d) @ c2_plus + diag_exp_batch(k0 * big_Q2 * (z_1d - d)) @ c2_minus)

        Sy = W_21 @ (diag_exp_batch(-k0 * big_Q1 * z_1d) @ c1_plus + diag_exp_batch(k0 * big_Q1 * (z_1d - d)) @ c1_minus) \
              + W_22 @ (diag_exp_batch(-k0 * big_Q2 * z_1d) @ c2_plus + diag_exp_batch(k0 * big_Q2 * (z_1d - d)) @ c2_minus)

        Ux = V_11 @ (-diag_exp_batch(-k0 * big_Q1 * z_1d) @ c1_plus + diag_exp_batch(k0 * big_Q1 * (z_1d - d)) @ c1_minus) \
              + V_12 @ (-diag_exp_batch(-k0 * big_Q2 * z_1d) @ c2_plus + diag_exp_batch(k0 * big_Q2 * (z_1d - d)) @ c2_minus)

        Uy = V_21 @ (-diag_exp_batch(-k0 * big_Q1 * z_1d) @ c1_plus + diag_exp_batch(k0 * big_Q1 * (z_1d - d)) @ c1_minus) \
              + V_22 @ (-diag_exp_batch(-k0 * big_Q2 * z_1d) @ c2_plus + diag_exp_batch(k0 * big_Q2 * (z_1d - d)) @ c2_minus)

        Sz = -1j * E_conv_i @ (Kx @ Uy - Ky @ Ux)
        Uz = -1j * (Kx @ Sy - Ky @ Sx)

        x_1d = jnp.arange(res_x, dtype=type_float).reshape((1, -1, 1))
        y_1d = jnp.arange(res_y, dtype=type_float).reshape((-1, 1, 1))

        x_1d = -1j * x_1d * period[0] / res_x
        y_1d = -1j * y_1d * period[1] / res_y

        x_2d = jnp.tile(x_1d, (res_y, 1, 1))
        y_2d = jnp.tile(y_1d, (1, res_x, 1))

        x_2d = x_2d * kx_vector
        y_2d = y_2d * ky_vector

        x_2d = x_2d.reshape((res_y, res_x, 1, len(kx_vector)))
        y_2d = y_2d.reshape((res_y, res_x, len(ky_vector), 1))

        exp_K = jnp.exp(x_2d) * jnp.exp(y_2d)
        exp_K = exp_K.reshape((res_y, res_x, -1))

        Ex = exp_K[:, :, None, :] @ Sx[:, None, None, :, :]
        Ey = exp_K[:, :, None, :] @ Sy[:, None, None, :, :]
        Ez = exp_K[:, :, None, :] @ Sz[:, None, None, :, :]

        Hx = -1j * exp_K[:, :, None, :] @ Ux[:, None, None, :, :]
        Hy = -1j * exp_K[:, :, None, :] @ Uy[:, None, None, :, :]
        Hz = -1j * exp_K[:, :, None, :] @ Uz[:, None, None, :, :]

        val = jnp.concatenate(
            (Ex.squeeze(-1), Ey.squeeze(-1), Ez.squeeze(-1), Hx.squeeze(-1), Hy.squeeze(-1), Hz.squeeze(-1)),
            axis=-1)

        field_cell = field_cell.at[res_z * idx_layer:res_z * (idx_layer + 1)].set(val)

        T_layer = big_A_i @ big_X @ T_layer

    return field_cell


def field_dist_1d_vanilla(wavelength, kx_vector, T1, layer_info_list, period,
                          pol, res_x=20, res_y=20, res_z=20, type_complex=jnp.complex128):

    k0 = 2 * jnp.pi / wavelength
    Kx = jnp.diag(kx_vector / k0)

    field_cell = jnp.zeros((res_z * len(layer_info_list), res_y, res_x, 3), dtype=type_complex)

    T_layer = T1

    # From the first layer
    for idx_layer, (E_conv_i, q, W, X, a_i, b, d) in enumerate(layer_info_list[::-1]):

        c1 = T_layer[:, None]
        c2 = b @ a_i @ X @ T_layer[:, None]

        Q = jnp.diag(q)

        if pol == 0:
            V = W @ Q
            EKx = None

        else:
            V = E_conv_i @ W @ Q
            EKx = E_conv_i @ Kx

        for k in range(res_z):
            z = k / res_z * d

            if pol == 0:
                Sy = W @ (diag_exp(-k0 * Q * z) @ c1 + diag_exp(k0 * Q * (z - d)) @ c2)
                Ux = V @ (-diag_exp(-k0 * Q * z) @ c1 + diag_exp(k0 * Q * (z - d)) @ c2)
                C = Kx @ Sy

                for j in range(res_y):
                    for i in range(res_x):
                        x = i * period[0] / res_x

                        Ey = Sy.T @ jnp.exp(-1j * kx_vector.reshape((-1, 1)) * x)
                        Hx = -1j * Ux.T @ jnp.exp(-1j * kx_vector.reshape((-1, 1)) * x)
                        Hz = -1j * C.T @ jnp.exp(-1j * kx_vector.reshape((-1, 1)) * x)

                        field_cell = field_cell.at[res_z * idx_layer + k, j, i].set([Ey[0, 0], Hx[0, 0], Hz[0, 0]])

            else:
                Uy = W @ (diag_exp(-k0 * Q * z) @ c1 + diag_exp(k0 * Q * (z - d)) @ c2)
                Sx = V @ (-diag_exp(-k0 * Q * z) @ c1 + diag_exp(k0 * Q * (z - d)) @ c2)

                C = EKx @ Uy  # there is a better option for convergence
                for j in range(res_y):
                    for i in range(res_x):
                        x = i * period[0] / res_x

                        Hy = Uy.T @ jnp.exp(-1j * kx_vector.reshape((-1, 1)) * x)
                        Ex = 1j * Sx.T @ jnp.exp(-1j * kx_vector.reshape((-1, 1)) * x)
                        Ez = (-1j) * C.T @ jnp.exp(-1j * kx_vector.reshape((-1, 1)) * x)

                        field_cell = field_cell.at[res_z * idx_layer + k, j, i].set([Hy[0, 0], Ex[0, 0], Ez[0, 0]])

        T_layer = a_i @ X @ T_layer

    return field_cell


def field_dist_1d_conical_vanilla(wavelength, kx_vector, n_I, theta, phi, T1, layer_info_list, period,
                                  res_x=20, res_y=20, res_z=20, type_complex=jnp.complex128):

    k0 = 2 * jnp.pi / wavelength
    ky = k0 * n_I * jnp.sin(theta) * jnp.sin(phi)
    Kx = jnp.diag(kx_vector / k0)

    field_cell = jnp.zeros((res_z * len(layer_info_list), res_y, res_x, 6), dtype=type_complex)

    T_layer = T1

    big_I = jnp.eye((len(T1))).astype(type_complex)

    # From the first layer
    for idx_layer, [E_conv_i, q_1, q_2, W_1, W_2, V_11, V_12, V_21, V_22, big_X, big_A_i, big_B, d] \
            in enumerate(layer_info_list[::-1]):

        c = jnp.block([[big_I], [big_B @ big_A_i @ big_X]]) @ T_layer

        ff = len(c) // 4

        c1_plus = c[0 * ff:1 * ff]
        c2_plus = c[1 * ff:2 * ff]
        c1_minus = c[2 * ff:3 * ff]
        c2_minus = c[3 * ff:4 * ff]

        big_Q1 = jnp.diag(q_1)
        big_Q2 = jnp.diag(q_2)

        for k in range(res_z):
            z = k / res_z * d

            Sx = W_2 @ (diag_exp(-k0 * big_Q2 * z) @ c2_plus + diag_exp(k0 * big_Q2 * (z - d)) @ c2_minus)
            Sy = V_11 @ (diag_exp(-k0 * big_Q1 * z) @ c1_plus + diag_exp(k0 * big_Q1 * (z - d)) @ c1_minus) \
                 + V_12 @ (diag_exp(-k0 * big_Q2 * z) @ c2_plus + diag_exp(k0 * big_Q2 * (z - d)) @ c2_minus)
            Ux = W_1 @ (-diag_exp(-k0 * big_Q1 * z) @ c1_plus + diag_exp(k0 * big_Q1 * (z - d)) @ c1_minus)
            Uy = V_21 @ (-diag_exp(-k0 * big_Q1 * z) @ c1_plus + diag_exp(k0 * big_Q1 * (z - d)) @ c1_minus) \
                 + V_22 @ (-diag_exp(-k0 * big_Q2 * z) @ c2_plus + diag_exp(k0 * big_Q2 * (z - d)) @ c2_minus)
            Sz = -1j * E_conv_i @ (Kx @ Uy - ky * Ux)
            Uz = -1j * (Kx @ Sy - ky * Sx)

            for j in range(res_y):
                for i in range(res_x):
                    # val = x_loop_1d_conical(period, res_x, kx_vector, Sx, Sy, Sz, Ux, Uy, Uz, i)
                    x = i * period[0] / res_x

                    exp_K = jnp.exp(-1j * kx_vector.reshape((-1, 1)) * x)
                    # exp_K = exp_K.flatten()

                    Ex = Sx.T @ exp_K
                    Ey = Sy.T @ exp_K
                    Ez = Sz.T @ exp_K

                    Hx = -1j * Ux.T @ exp_K
                    Hy = -1j * Uy.T @ exp_K
                    Hz = -1j * Uz.T @ exp_K

                    val = [Ex[0, 0], Ey[0, 0], Ez[0, 0], Hx[0, 0], Hy[0, 0], Hz[0, 0]]
                    field_cell = field_cell.at[res_z * idx_layer + k, j, i].set(val)

        T_layer = big_A_i @ big_X @ T_layer

    return field_cell


def field_dist_2d_vanilla(wavelength, kx_vector, n_I, theta, phi, fourier_order_x, fourier_order_y, T1, layer_info_list, period,
                          res_x=20, res_y=20, res_z=20, type_complex=jnp.complex128):

    k0 = 2 * jnp.pi / wavelength

    fourier_indices_y = jnp.arange(-fourier_order_y, fourier_order_y + 1)
    ff_x = fourier_order_x * 2 + 1
    ff_y = fourier_order_y * 2 + 1
    ky_vector = k0 * (n_I * jnp.sin(theta) * jnp.sin(phi) + fourier_indices_y * (
            wavelength / period[1])).astype(type_complex)

    Kx = jnp.diag(jnp.tile(kx_vector, ff_y).flatten()) / k0
    Ky = jnp.diag(jnp.tile(ky_vector.reshape((-1, 1)), ff_x).flatten()) / k0

    field_cell = jnp.zeros((res_z * len(layer_info_list), res_y, res_x, 6), dtype=type_complex)

    T_layer = T1

    big_I = jnp.eye((len(T1))).astype(type_complex)

    # From the first layer
    for idx_layer, (E_conv_i, q, W_11, W_12, W_21, W_22, V_11, V_12, V_21, V_22, big_X, big_A_i, big_B, d)\
            in enumerate(layer_info_list[::-1]):
        c = jnp.block([[big_I], [big_B @ big_A_i @ big_X]]) @ T_layer

        ff = len(c) // 4

        c1_plus = c[0 * ff:1 * ff]
        c2_plus = c[1 * ff:2 * ff]
        c1_minus = c[2 * ff:3 * ff]
        c2_minus = c[3 * ff:4 * ff]

        q1 = q[:len(q) // 2]
        q2 = q[len(q) // 2:]
        big_Q1 = jnp.diag(q1)
        big_Q2 = jnp.diag(q2)

        for k in range(res_z):
            z = k / res_z * d

            Sx = W_11 @ (diag_exp(-k0 * big_Q1 * z) @ c1_plus + diag_exp(k0 * big_Q1 * (z - d)) @ c1_minus) \
                 + W_12 @ (diag_exp(-k0 * big_Q2 * z) @ c2_plus + diag_exp(k0 * big_Q2 * (z - d)) @ c2_minus)
            Sy = W_21 @ (diag_exp(-k0 * big_Q1 * z) @ c1_plus + diag_exp(k0 * big_Q1 * (z - d)) @ c1_minus) \
                 + W_22 @ (diag_exp(-k0 * big_Q2 * z) @ c2_plus + diag_exp(k0 * big_Q2 * (z - d)) @ c2_minus)
            Ux = V_11 @ (-diag_exp(-k0 * big_Q1 * z) @ c1_plus + diag_exp(k0 * big_Q1 * (z - d)) @ c1_minus) \
                 + V_12 @ (-diag_exp(-k0 * big_Q2 * z) @ c2_plus + diag_exp(k0 * big_Q2 * (z - d)) @ c2_minus)
            Uy = V_21 @ (-diag_exp(-k0 * big_Q1 * z) @ c1_plus + diag_exp(k0 * big_Q1 * (z - d)) @ c1_minus) \
                 + V_22 @ (-diag_exp(-k0 * big_Q2 * z) @ c2_plus + diag_exp(k0 * big_Q2 * (z - d)) @ c2_minus)
            Sz = -1j * E_conv_i @ (Kx @ Uy - Ky @ Ux)
            Uz = -1j * (Kx @ Sy - Ky @ Sx)

            for j in range(res_y):
                y = j * period[1] / res_y
                for i in range(res_x):
                    x = i * period[0] / res_x

                    exp_K = jnp.exp(-1j * kx_vector.reshape((1, -1)) * x) * jnp.exp(
                        -1j * ky_vector.reshape((-1, 1)) * y)
                    exp_K = exp_K.flatten()

                    Ex = Sx.T @ exp_K
                    Ey = Sy.T @ exp_K
                    Ez = Sz.T @ exp_K
                    Hx = -1j * Ux.T @ exp_K
                    Hy = -1j * Uy.T @ exp_K
                    Hz = -1j * Uz.T @ exp_K

                    val = [Ex[0], Ey[0], Ez[0], Hx[0], Hy[0], Hz[0]]

                    field_cell = field_cell.at[res_z * idx_layer + k, j, i].set(val)
        T_layer = big_A_i @ big_X @ T_layer

    return field_cell


# def field_dist_2d_lax(wavelength, kx_vector, n_I, theta, phi, fourier_order_x, fourier_order_y, T1, layer_info_list, period,
#                   resolution=(10, 10, 10),
#                   type_complex=jnp.complex128):
#
#     k0 = 2 * jnp.pi / wavelength
#     fourier_indices_y = jnp.arange(-fourier_order_y, fourier_order_y + 1)
#     ff_x = fourier_order_x * 2 + 1
#     ff_y = fourier_order_y * 2 + 1
#     ff_xy = ff_x * ff_y
#     ky_vector = k0 * (n_I * jnp.sin(theta) * jnp.sin(phi) + fourier_indices_y * (
#             wavelength / period[1])).astype(type_complex)
#
#     Kx = jnp.diag(jnp.tile(kx_vector, ff_y).flatten()) / k0
#     Ky = jnp.diag(jnp.tile(ky_vector.reshape((-1, 1)), ff_x).flatten()) / k0
#
#     resolution_x, resolution_y, resolution_z = resolution
#     field_cell = jnp.zeros((resolution_z * len(layer_info_list), resolution_y, resolution_x, 6), dtype=type_complex)
#
#     T_layer = T1
#
#     big_I = jnp.eye((len(T1))).astype(type_complex)
#
#     # From the first layer
#     for idx_layer, (E_conv_i, q, W_11, W_12, W_21, W_22, V_11, V_12, V_21, V_22, big_X, big_A_i, big_B, d)\
#             in enumerate(layer_info_list[::-1]):
#
#         c = jnp.block([[big_I], [big_B @ big_A_i @ big_X]]) @ T_layer
#         i=j=k=0  # delete
#
#         args = [k, j, i, k0, resolution_z, resolution_y, resolution_x, idx_layer, d, kx_vector, ky_vector, q, period, c, Kx, Ky, E_conv_i, W_11, W_12, W_21, W_22, V_11, V_12, V_21]
#
#         res_size = 1 * 9 + ff_x + ff_y + 2*ff_xy + 2 + 2*2*ff_xy + 11 * ff_xy**2
#         res = jnp.zeros(res_size, dtype=type_complex)
#
#         b = 0
#         for ix, item in enumerate(args):
#             if type(item) in (float, int):
#                 length = 1
#                 val = item
#             elif isinstance(item, jax.numpy.ndarray):
#                 length = item.size
#                 val = item.flatten()
#             elif type(item) is list:
#                 length = len(item)
#                 val = jnp.array(item)
#             else:
#                 raise
#
#             res[b:b+length] = val
#             b += length
#
#         ress = jnp.tile(res, (resolution_z * resolution_y * resolution_x, 1))
#
#         base = jnp.arange(resolution_x)
#         i_list = jnp.tile(base, resolution_z * resolution_y)[:,None]
#         j_list = jnp.tile(jnp.repeat(base, resolution_x), resolution_z)[:,None]
#         k_list = jnp.repeat(base, resolution_x * resolution_y)[:,None]
#         kji = jnp.hstack((k_list, j_list, i_list))
#
#         resss = ress.at[:, :3].set(kji)
#
#         def calc(field_cell, args):
#             k = args[0]
#             j = args[1]
#             i = args[2]
#             k0 = args[3]
#             resolution_z = args[4]
#             resolution_y = args[5]
#             resolution_x = args[6]
#             idx_layer = args[7]
#             d = args[8]
#
#             b, e = 9, 9 + ff_x
#             kx_vector = args[b:e]
#             b, e = e, e + ff_y
#             ky_vector = args[b:e]
#             b, e = e, e + 2 * ff_xy
#             q = args[b:e]
#             b, e = e, e + 2
#             period = args[b:e]
#
#             b, e = e, e + 2 * 2 * ff_xy
#             c = args[b:e].reshape((-1, 1))
#
#             b, e = e, e + ff_xy * ff_xy
#             Kx = args[b:e].reshape((ff_xy, ff_xy))
#             b, e = e, e + ff_xy * ff_xy
#             Ky = args[b:e].reshape((ff_xy, ff_xy))
#             b, e = e, e + ff_xy * ff_xy
#             E_conv_i = args[b:e].reshape((ff_xy, ff_xy))
#             b, e = e, e + ff_xy * ff_xy
#             W_11 = args[b:e].reshape((ff_xy, ff_xy))
#             b, e = e, e + ff_xy * ff_xy
#             W_12 = args[b:e].reshape((ff_xy, ff_xy))
#             b, e = e, e + ff_xy * ff_xy
#             W_21 = args[b:e].reshape((ff_xy, ff_xy))
#             b, e = e, e + ff_xy * ff_xy
#             W_22 = args[b:e].reshape((ff_xy, ff_xy))
#             b, e = e, e + ff_xy * ff_xy
#             V_11 = args[b:e].reshape((ff_xy, ff_xy))
#             b, e = e, e + ff_xy * ff_xy
#             V_12 = args[b:e].reshape((ff_xy, ff_xy))
#             b, e = e, e + ff_xy * ff_xy
#             V_21 = args[b:e].reshape((ff_xy, ff_xy))
#             b, e = e, e + ff_xy * ff_xy
#             V_22 = args[b:e].reshape((ff_xy, ff_xy))
#
#             y = j * period[1] / resolution_y
#             Sx, Sy, Ux, Uy, Sz, Uz = z_loop_2d(k, c, k0, Kx, Ky, resolution_z, E_conv_i, q, W_11, W_12, W_21, W_22,
#                                                V_11, V_12, V_21, V_22, d)
#             val = x_loop_2d(period, resolution_x, kx_vector, ky_vector, Sx, Sy, Sz, Ux, Uy, Uz, y, i)
#             field_cell = field_cell.at[(resolution_z * idx_layer + k).real.astype(int), j.real.astype(int), i.real.astype(int)].set(val)
#
#             return field_cell, val
#
#         field_cell, _ = jax.lax.scan(calc, field_cell, resss)
#     return field_cell
#
#
#
# def field_dist_2d_lax_heavy(wavelength, kx_vector, n_I, theta, phi, fourier_order_x, fourier_order_y, T1, layer_info_list, period,
#                   resolution=(10, 10, 10),
#                   type_complex=jnp.complex128):
#
#     k0 = 2 * jnp.pi / wavelength
#     fourier_indices_y = jnp.arange(-fourier_order_y, fourier_order_y + 1)
#     ff_x = fourier_order_x * 2 + 1
#     ff_y = fourier_order_y * 2 + 1
#     ff_xy = ff_x * ff_y
#     ky_vector = k0 * (n_I * jnp.sin(theta) * jnp.sin(phi) + fourier_indices_y * (
#             wavelength / period[1])).astype(type_complex)
#
#     Kx = jnp.diag(jnp.tile(kx_vector, ff_y).flatten()) / k0
#     Ky = jnp.diag(jnp.tile(ky_vector.reshape((-1, 1)), ff_x).flatten()) / k0
#
#     resolution_x, resolution_y, resolution_z = resolution
#     field_cell = jnp.zeros((resolution_z * len(layer_info_list), resolution_y, resolution_x, 6), dtype=type_complex)
#
#     T_layer = T1
#
#     big_I = jnp.eye((len(T1))).astype(type_complex)
#
#     # From the first layer
#     for idx_layer, (E_conv_i, q, W_11, W_12, W_21, W_22, V_11, V_12, V_21, V_22, big_X, big_A_i, big_B, d)\
#             in enumerate(layer_info_list[::-1]):
#
#         c = jnp.block([[big_I], [big_B @ big_A_i @ big_X]]) @ T_layer
#         i=j=k=0  # delete
#
#         args = [k, j, i, k0, resolution_z, resolution_y, resolution_x, idx_layer, d, kx_vector, ky_vector, q, period, c, Kx, Ky, E_conv_i, W_11, W_12, W_21, W_22, V_11, V_12, V_21]
#
#         res_size = 1 * 9 + ff_x + ff_y + 2*ff_xy + 2 + 2*2*ff_xy + 11 * ff_xy**2
#         res = jnp.zeros(res_size, dtype=type_complex)
#
#         b = 0
#         for ix, item in enumerate(args):
#             if type(item) in (float, int):
#                 length = 1
#                 val = item
#             elif isinstance(item, jax.numpy.ndarray):
#                 length = item.size
#                 val = item.flatten()
#             elif type(item) is list:
#                 length = len(item)
#                 val = jnp.array(item)
#             else:
#                 raise
#
#             res[b:b+length] = val
#             b += length
#
#         ress = jnp.tile(res, (resolution_z * resolution_y * resolution_x, 1))
#
#         base = jnp.arange(resolution_x)
#         i_list = jnp.tile(base, resolution_z * resolution_y)[:,None]
#         j_list = jnp.tile(jnp.repeat(base, resolution_x), resolution_z)[:,None]
#         k_list = jnp.repeat(base, resolution_x * resolution_y)[:,None]
#         kji = jnp.hstack((k_list, j_list, i_list))
#
#         resss = ress.at[:, :3].set(kji)
#
#         def calc(field_cell, args):
#             k = args[0]
#             j = args[1]
#             i = args[2]
#             k0 = args[3]
#             resolution_z = args[4]
#             resolution_y = args[5]
#             resolution_x = args[6]
#             idx_layer = args[7]
#             d = args[8]
#
#             b, e = 9, 9 + ff_x
#             kx_vector = args[b:e]
#             b, e = e, e + ff_y
#             ky_vector = args[b:e]
#             b, e = e, e + 2 * ff_xy
#             q = args[b:e]
#             b, e = e, e + 2
#             period = args[b:e]
#
#             b, e = e, e + 2 * 2 * ff_xy
#             c = args[b:e].reshape((-1, 1))
#
#             b, e = e, e + ff_xy * ff_xy
#             Kx = args[b:e].reshape((ff_xy, ff_xy))
#             b, e = e, e + ff_xy * ff_xy
#             Ky = args[b:e].reshape((ff_xy, ff_xy))
#             b, e = e, e + ff_xy * ff_xy
#             E_conv_i = args[b:e].reshape((ff_xy, ff_xy))
#             b, e = e, e + ff_xy * ff_xy
#             W_11 = args[b:e].reshape((ff_xy, ff_xy))
#             b, e = e, e + ff_xy * ff_xy
#             W_12 = args[b:e].reshape((ff_xy, ff_xy))
#             b, e = e, e + ff_xy * ff_xy
#             W_21 = args[b:e].reshape((ff_xy, ff_xy))
#             b, e = e, e + ff_xy * ff_xy
#             W_22 = args[b:e].reshape((ff_xy, ff_xy))
#             b, e = e, e + ff_xy * ff_xy
#             V_11 = args[b:e].reshape((ff_xy, ff_xy))
#             b, e = e, e + ff_xy * ff_xy
#             V_12 = args[b:e].reshape((ff_xy, ff_xy))
#             b, e = e, e + ff_xy * ff_xy
#             V_21 = args[b:e].reshape((ff_xy, ff_xy))
#             b, e = e, e + ff_xy * ff_xy
#             V_22 = args[b:e].reshape((ff_xy, ff_xy))
#
#             y = j * period[1] / resolution_y
#             Sx, Sy, Ux, Uy, Sz, Uz = z_loop_2d(k, c, k0, Kx, Ky, resolution_z, E_conv_i, q, W_11, W_12, W_21, W_22,
#                                                V_11, V_12, V_21, V_22, d)
#             val = x_loop_2d(period, resolution_x, kx_vector, ky_vector, Sx, Sy, Sz, Ux, Uy, Uz, y, i)
#             field_cell = field_cell.at[(resolution_z * idx_layer + k).real.astype(int), j.real.astype(int), i.real.astype(int)].set(val)
#
#             return field_cell, val
#
#         field_cell, _ = jax.lax.scan(calc, field_cell, resss)
#     return field_cell


def field_plot(field_cell, pol=0, plot_indices=(1, 1, 1, 1, 1, 1), y_slice=0, z_slice=-1, zx=True, yx=True):
    try:
        import matplotlib.pyplot as plt
    except (ImportError, ModuleNotFoundError) as e:
        print(e)
        print('To use cal_field(), please install matplotlib')
        raise e

    if field_cell.shape[-1] == 6:  # 2D grating
        title = ['2D Ex', '2D Ey', '2D Ez', '2D Hx', '2D Hy', '2D Hz', ]
    else:  # 1D grating
        if pol == 0:  # TE
            title = ['1D Ey', '1D Hx', '1D Hz', ]
        else:  # TM
            title = ['1D Hy', '1D Ex', '1D Ez', ]

    if zx:
        for idx in range(len(title)):
            if plot_indices[idx]:
                plt.imshow((abs(field_cell[:, y_slice, :, idx]) ** 2), cmap='jet', aspect='auto')
                # plt.clim(0, 2)  # identical to caxis([-4,4]) in MATLAB
                plt.colorbar()
                plt.title(f'{title[idx]}, Side View')
                plt.xlabel('X')
                plt.ylabel('Z')
                plt.show()
    if yx:
        for idx in range(len(title)):
            if plot_indices[idx]:
                plt.imshow((abs(field_cell[z_slice, :, :, idx]) ** 2), cmap='jet', aspect='auto')
                # plt.clim(0, 3.5)  # identical to caxis([-4,4]) in MATLAB
                plt.colorbar()
                plt.title(f'{title[idx]}, Top View')
                plt.xlabel('X')
                plt.ylabel('Y')
                plt.show()


def diag_exp(x):
    return jnp.diag(jnp.exp(jnp.diag(x)))


def diag_exp_batch(x):
    res = jnp.zeros(x.shape, dtype=x.dtype)
    ix = jnp.diag_indices_from(x[0])
    res = res.at[:, ix[0], ix[1]].set(jnp.exp(x[:, ix[0], ix[1]]))
    return res
